from django.contrib import admin
from django.forms import models, ValidationError
from django.utils.translation import get_language, ugettext_lazy as _

# messy stuff
from l10n.l10n_settings import get_l10n_setting
from livesettings import config_value
from product.models import Category, CategoryTranslation, CategoryImage, CategoryImageTranslation, \
                                   OptionGroup, OptionGroupTranslation, Option, OptionTranslation, Product, \
                                   CustomProduct, CustomTextField, CustomTextFieldTranslation, ConfigurableProduct, \
                                   DownloadableProduct, SubscriptionProduct, Trial, ProductVariation, ProductAttribute, \
                                   Price, ProductImage, ProductImageTranslation, default_weight_unit, \
                                   default_dimension_unit, ProductTranslation, Discount, TaxClass, AttributeOption
from satchmo_utils.thumbnail.field import ImageWithThumbnailField
from satchmo_utils.thumbnail.widgets import AdminImageWithThumbnailWidget
import re

class CategoryTranslation_Inline(admin.StackedInline):
    model = CategoryTranslation
    extra = 1

class CategoryImage_Inline(admin.TabularInline):
    model = CategoryImage
    extra = 3
    
    formfield_overrides = {
        ImageWithThumbnailField : {'widget' : AdminImageWithThumbnailWidget},
    }

class CategoryImageTranslation_Inline(admin.StackedInline):
    model = CategoryImageTranslation
    extra = 1

class DiscountOptions(admin.ModelAdmin):
    list_display=('site', 'description','active')
    list_display_links = ('description',)
    raw_id_fields = ('validProducts',)

class OptionGroupTranslation_Inline(admin.StackedInline):
    model = OptionGroupTranslation
    extra = 1

class Option_Inline(admin.TabularInline):
    model = Option
    extra = 5

class OptionTranslation_Inline(admin.StackedInline):
    model = OptionTranslation
    extra = 1

class CustomTextField_Inline(admin.TabularInline):
    model = CustomTextField
    extra = 3

class CustomTextFieldTranslation_Inline(admin.StackedInline):
    model = CustomTextFieldTranslation
    extra = 1

class Trial_Inline(admin.StackedInline):
    model = Trial
    extra = 2

class ProductAttributeForm(models.ModelForm):
    
    def clean_validation(self):
        validation = self.cleaned_data['validation']
        try:
            re.compile(validation)
        except:
            raise ValidationError(_("Invalid regular expression"))
        return validation
        

class ProductAttributeAdmin(admin.ModelAdmin):
    form = ProductAttributeForm
    prepopulated_fields = {"name": ("description",)}
    

class ProductAttributeInlineForm(models.ModelForm):
    
    def clean_value(self):
        value = self.cleaned_data['value']
        attribute = self.cleaned_data['option']
        product = self.cleaned_data['product']
        function_name = attribute.validation.split('.')[-1]
        import_name = '.'.join(attribute.validation.split('.')[:-1])
        import_module = __import__(import_name, fromlist=[function_name])
        validation_function = getattr(import_module, function_name)
        success, valid_value = validation_function(value, product)
        if not success:
            raise ValidationError(attribute.error_message)
        return valid_value

class ProductAttribute_Inline(admin.TabularInline):
    model = ProductAttribute
    extra = 2
    form = ProductAttributeInlineForm

class Price_Inline(admin.TabularInline):
    model = Price
    extra = 1

class ProductImage_Inline(admin.StackedInline):
    model = ProductImage
    extra = 3
    
    formfield_overrides = {
        ImageWithThumbnailField : {'widget' : AdminImageWithThumbnailWidget},
    }
 
class ProductTranslation_Inline(admin.TabularInline): 
    model = ProductTranslation 
    extra = 1    

class ProductImageTranslation_Inline(admin.StackedInline):
    model = ProductImageTranslation
    extra = 1

class CategoryAdminForm(models.ModelForm):
    
    def clean_parent(self):
        parent = self.cleaned_data['parent']
        slug = self.cleaned_data['slug']
        if parent and slug:
            if parent.slug == slug:
                raise ValidationError(_("You must not save a category in itself!"))

            for p in parent._recurse_for_parents(parent):
                if slug == p.slug:
                    raise ValidationError(_("You must not save a category in itself!"))
                    
        return parent

class CategoryOptions(admin.ModelAdmin):
    
    if config_value('SHOP','SHOW_SITE'):
        list_display = ('site',)
    else:
        list_display = ()
    
    list_display += ('name', '_parents_repr', 'is_active')
    list_display_links = ('name',)
    ordering = ['site', 'parent__id', 'ordering', 'name']
    #inlines = [CategoryImage_Inline]
    #if get_l10n_setting('show_translations'):
    #    inlines.append(CategoryTranslation_Inline)
    filter_horizontal = ('related_categories',)
    form = CategoryAdminForm

    actions = ('mark_active', 'mark_inactive')

    def mark_active(self, request, queryset):
        queryset.update(is_active=True)

    def mark_inactive(self, request, queryset):
        queryset.update(is_active=False)

class CategoryImageOptions(admin.ModelAdmin):
    inlines = [CategoryImageTranslation_Inline]

class OptionGroupOptions(admin.ModelAdmin):
    inlines = [Option_Inline]
    if get_l10n_setting('show_translations'):
        inlines.append(OptionGroupTranslation_Inline)
    if config_value('SHOP','SHOW_SITE'):
        list_display = ('site',)
    else:
        list_display = ()
    list_display += ('name',)

class OptionOptions(admin.ModelAdmin):
    inlines = []
    if get_l10n_setting('show_translations'):
        inlines.append(OptionTranslation_Inline)

class ProductOptions(admin.ModelAdmin):
    
    def make_active(self, request, queryset):
        rows_updated = queryset.update(active=True)
        if rows_updated == 1:
            message_bit = _("1 product was")
        else:
            message_bit = _("%s products were" % rows_updated)
        self.message_user(request, _("%s successfully marked as active") % message_bit)
    make_active.short_description = _("Mark selected products as active")
    
    def make_inactive(self, request, queryset):
        rows_updated = queryset.update(active=False)
        if rows_updated == 1:
            message_bit = _("1 product was")
        else:
            message_bit = _("%s products were" % rows_updated)
        self.message_user(request, _("%s successfully marked as inactive") % message_bit)
    make_inactive.short_description = _("Mark selected products as inactive")
    
    def make_featured(self, request, queryset):
        rows_updated = queryset.update(featured=True)
        if rows_updated == 1:
            message_bit = _("1 product was")
        else:
            message_bit = _("%s products were" % rows_updated)
        self.message_user(request, _("%s successfully marked as featured") % message_bit)
    make_featured.short_description = _("Mark selected products as featured")
    
    def make_unfeatured(self, request, queryset):
        rows_updated = queryset.update(featured=False)
        if rows_updated == 1:
            message_bit = _("1 product was")
        else:
            message_bit = _("%s products were" % rows_updated)
        self.message_user(request, _("%s successfully marked as not featured") % message_bit)
    make_unfeatured.short_description = _("Mark selected products as not featured")

    if config_value('SHOP','SHOW_SITE'):
        list_display = ('site',)
    else:
        list_display = ()

    list_display += ('slug', 'name', 'unit_price', 'items_in_stock', 'active','featured', 'get_subtypes')
    list_display_links = ('slug', 'name')
    list_filter = ('category', 'date_added','active','featured')
    actions = ('make_active', 'make_inactive', 'make_featured', 'make_unfeatured')
    fieldsets = (
    (None, {'fields': ('site', 'category', 'name', 'slug', 'sku', 'description', 'short_description', 'date_added', 
            'active', 'featured', 'items_in_stock','total_sold','ordering', 'shipclass')}), (_('Meta Data'), {'fields': ('meta',), 'classes': ('collapse',)}), 
            (_('Item Dimensions'), {'fields': (('length', 'length_units','width','width_units','height','height_units'),('weight','weight_units')), 'classes': ('collapse',)}), 
            (_('Tax Options'), {'fields':('taxable', 'taxClass'), 'classes': ('collapse',)}), 
            (_('Related Products'), {'fields':('related_items','also_purchased'),'classes':('collapse',)}), )
    search_fields = ['slug', 'sku', 'name']
    #inlines = [ProductAttribute_Inline, Price_Inline, ProductImage_Inline]
    inlines = [Price_Inline]
    #if get_l10n_setting('show_translations'):
    #    inlines.append(ProductTranslation_Inline)
    filter_horizontal = ('category',)
    
    def formfield_for_dbfield(self, db_field, **kwargs):
        field = super(ProductOptions, self).formfield_for_dbfield(db_field, **kwargs)
        fieldname = db_field.name
        if fieldname in ("length_units", "width_units", "height_units"): 
            field.initial = default_dimension_unit()
        elif fieldname == "weight_units":
            field.initial = default_weight_unit()
        return field

class CustomProductOptions(admin.ModelAdmin):
    #inlines = [CustomTextField_Inline]
    inlines = []

class CustomTextFieldOptions(admin.ModelAdmin):
    inlines = []
    if get_l10n_setting('show_translations'):
        inlines.append(CustomTextFieldTranslation_Inline)

class SubscriptionProductOptions(admin.ModelAdmin):
    inlines = [Trial_Inline]

class ProductVariationOptions(admin.ModelAdmin):
    filter_horizontal = ('options',)

class ProductImageOptions(admin.ModelAdmin):
    inlines = []
    if get_l10n_setting('show_translations'):
        inlines.append(ProductImageTranslation_Inline)

admin.site.register(Category, CategoryOptions)
admin.site.register(CategoryImage, CategoryImageOptions)
admin.site.register(Discount, DiscountOptions)
admin.site.register(OptionGroup, OptionGroupOptions)
admin.site.register(Option, OptionOptions)
admin.site.register(Product, ProductOptions)
admin.site.register(CustomProduct, CustomProductOptions)
admin.site.register(CustomTextField, CustomTextFieldOptions)
admin.site.register(ConfigurableProduct)
admin.site.register(DownloadableProduct)
admin.site.register(SubscriptionProduct, SubscriptionProductOptions)
admin.site.register(ProductVariation, ProductVariationOptions)
admin.site.register(TaxClass)
admin.site.register(AttributeOption, ProductAttributeAdmin)
admin.site.register(ProductImage, ProductImageOptions)




#!/usr/bin/env python
"""Base model classes for django-hotsauce"""
import logging
log = logging.getLogger(__name__)

from notmm.dbapi.orm import RelationProxy

__all__ = ['ModelManager']

class ModelManager(object):

    model = None # Category

    def __init__(self, connection=None, model=None, **kwargs):
        self.db = connection
        self.kwargs = kwargs
        if model is not None:
            self.model = model
        if connection is not None:
            self.objects = RelationProxy(getattr(connection, self.model))
        else:
            self.objects = None

    def __str__(self):
        return "<ModelManager: %s>" % self.model
   
    def find_or_create(self, **kwargs):
        try:
            obj = getattr(self.db, self.model).findone(**kwargs)
            if obj is not None:
                log.debug("Found object: %s!!!" % obj)
        except:
            # create
            tx = self.objects.t.create(**kwargs)
            self.db.execute(tx)
            self.db.commit()
            log.debug("Created new object: %s" % tx)
            obj = self.objects.get(**kwargs)
        return obj

    def save(self, commit=True):
        lock = self.db.write_lock()
        with lock:
            #if self.initialized:
            tx = self.extent.t.create(**self.kwargs)
            if commit:
                self.db.execute(tx)
                self.db._commit()
        lock.release()
        return self




"""
Urls for product feeds, note that this does not have to get added manually to the urls, it will be added automatically by satchmo core if this app is installed.
"""
from django.conf.urls.defaults import *
import logging
log = logging.getLogger('product_feeds.urls')

urlpatterns = patterns('satchmo_ext.product_feeds.views',
    (r'atom/$', 'product_feed', {}, 'satchmo_atom_feed'),
    (r'atom/(?P<category>\w+)/$', 'product_feed', {}, 'satchmo_atom_category_feed'),
    (r'products.csv$', 'admin_product_feed', {'template' : "product_feeds/product_feed.csv"}, 'satchmo_product_feed'),
)

feedpatterns = patterns('',
    (r'^feed/', include('satchmo_ext.product_feeds.urls'))
)

def add_feed_urls(sender, patterns=(), **kwargs):
    log.debug("Adding product_feed urls")
    patterns += feedpatterns



from livesettings import *
from django.utils.translation import ugettext_lazy as _

gettext = lambda s:s
_strings = (gettext('CreditCard'), gettext('Credit Card'))

PAYMENT_GROUP = ConfigurationGroup('PAYMENT_TRUSTCOMMERCE', 
    _('TrustCommerce Payment Settings'), 
    ordering=102)

config_register_list(

    StringValue(PAYMENT_GROUP,
        'KEY',
        description=_("Module key"),
        hidden=True,
        default = 'TRUSTCOMMERCE'),

    ModuleValue(PAYMENT_GROUP,
        'MODULE',
        description=_('Implementation module'),
        hidden=True,
        default = 'payment.modules.trustcommerce'),
        
    BooleanValue(PAYMENT_GROUP, 
        'SSL', 
        description=_("Use SSL for the checkout pages?"), 
        default=False),
    
    BooleanValue(PAYMENT_GROUP, 
        'AVS', 
        description=_("Use Address Verification System (AVS)?"), 
        default=False),
    
    BooleanValue(PAYMENT_GROUP, 
        'LIVE', 
        description=_("Accept real payments"),
        help_text=_("False if you want to be in test mode"),
        default=False),
    
    StringValue(PAYMENT_GROUP,
        'AUTH_TYPE',
        description=_("Type of authorization to perform."),
        help_text = _("Refer to manual for details on the different types."),
        default = 'sale',
        choices = [('sale', _('Sale')),
                    ('preauth', _('Preauth'))]
        ),
        
    StringValue(PAYMENT_GROUP,
        'LABEL',
        description=_('English name for this group on the checkout screens'),
        default = 'Credit Cards',
        help_text = _('This will be passed to the translation utility')),

    StringValue(PAYMENT_GROUP,
        'URL_BASE',
        description=_('The url base used for constructing urlpatterns which will use this module'),
        default = r'^credit/'),

    MultipleStringValue(PAYMENT_GROUP,
        'CREDITCHOICES',
        description=_('Available credit cards'),
        choices = (
            (('American Express', 'American Express')),
            (('Visa','Visa')),
            (('Mastercard','Mastercard')),
            (('Discover','Discover'))),
        default = ('Visa', 'Mastercard', 'Discover')),
    
    StringValue(PAYMENT_GROUP, 
        'LOGIN', 
        description=_('Your TrustCommerce login'),
        default=""),
    
    StringValue(PAYMENT_GROUP, 
        'PASSWORD', 
        description=_('Your TrustCommerce password'),
        default=""),
        
    BooleanValue(PAYMENT_GROUP,
        'EXTRA_LOGGING',
        description=_("Verbose logs"),
        help_text=_("Add extensive logs during post."),
        default=False)
)



#!/usr/bin/env python

from test_support import unittest

# schevo
#from schevo.database2 import Database 
from notmm.dbapi.orm import ZODBFileStorageProxy
from model import AuthorManager, MessageManager


connection = ZODBFileStorageProxy('127.0.0.1:4343')

class TutorialTestCase(unittest.TestCase):

    def setUp(self):
        
        super(TutorialTestCase, self).setUp()
        self.db = MessageManager(connection).db_connection

        # populate the database with sample data?
        # self.db.populate()

    def tearDown(self):
        self.db.close()
    
    def test_open_database(self):

        db = self.db
        #self.assertEqual(repr(db), "<Database u'Schevo Database' :: V 25>")
        
        # now lets make some introspection on db 
        self.assertEqual(isinstance(db, ZODBFileStorageProxy), True, type(db))
        #self.assertEqual(db.version, 27) 
        self.assertEqual(db.format==2, True, db.format)

    def test_find_author(self):
        import schevo.query as Q
        Person = self.db.Author.findone(username=u"Etienne Robillard")
        #self.failUnlessEqual(isinstance(Person, db.Actor.__class__), True, type(Person))
        blogentries = self.db.Message.find(author=Person)
        self.assertEqual(isinstance(blogentries, Q.ResultsList), True)
        #print blogentries



from livesettings import config_register, StringValue, IntegerValue, BooleanValue, MultipleStringValue
from satchmo_store.shop.config import SHOP_GROUP

from django.utils.translation import ugettext_lazy as _

config_register(
    BooleanValue(SHOP_GROUP,
    'AUTHENTICATION_REQUIRED',
    description=_("Only authenticated users can check out"),
    help_text=_("Users will be required to authenticate (and create an account if neccessary) before checkout."),
    default=False,
    )
)

config_register(
    MultipleStringValue(SHOP_GROUP,
    'REQUIRED_BILLING_DATA',
    description=_("Required billing data"),
    help_text=_(
        "Users may be required to provide some set of billing address. Other fields are optional. "
        "You may shorten the checkout process here, but be careful, as this may leave you orders "
        "with almost no customer data! Some payment modules may override this setting."
        ),
    default=('email', 'first_name', 'last_name', 'phone', 'street1', 'city', 'postal_code', 'country'),
    choices=(
        ('email', _("Email")),
        ('title', _("Title")),
        ('first_name', _("First name")),
        ('last_name', _("Last name")),
        ('phone', _("Phone")),
        ('addressee', _("Addressee")),
        ('organization', _("Organization")),
        ('street1', _("Street")),
        ('street2', _("Street (second line)")),
        ('city', _("City")),
        ('state', _("State/Province")),
        ('postal_code', _("Postal code/ZIP")),
        ('country', _("Country"))
        )
    )
)
# I am doing it this way instead of a boolean for email verification because I
# intend to add a "manual approval" style of account verification. -Bruce
ACCOUNT_VERIFICATION = config_register(StringValue(SHOP_GROUP,
    'ACCOUNT_VERIFICATION',
    description=_("Account Verification"),
    help_text=_("Select the style of account verification.  'Immediate' means no verification needed."),
    default="IMMEDIATE",
    choices=[('IMMEDIATE', _('Immediate')),
             ('EMAIL', _('Email'))]
    ))
             
config_register(
    IntegerValue(SHOP_GROUP,
    'ACCOUNT_ACTIVATION_DAYS',
    description=_('Days to verify account'),
    default=7,
    requires=ACCOUNT_VERIFICATION,
    requiresvalue='EMAIL')
)



VERSION = (1, 3, 3)

# Dynamically calculate the version based on VERSION tuple
if len(VERSION)>2 and VERSION[2] is not None:
    str_version = "%d.%d_%s" % VERSION[:3]
else:
    str_version = "%d.%d" % VERSION[:2]

__version__ = str_version



import logging
from product.models import ConfigurableProduct

class LogMiddleware(object):
    """
    This middleware setups a cache to store information on which items are
    viewed.
    """
    def process_view(self, request, view_func, view_args, view_kwargs):
        logger1 = logging.getLogger('stats')
        if 'slug' in view_kwargs and 'queryset' in view_kwargs:
            try:
                if isinstance(view_kwargs['queryset'][0], ConfigurableProduct):
                    logger1.info("Viewing item %s" % view_kwargs['slug'])
            except IndexError:
                pass
        return None



"""
Stores details about the available payment options.
Also stores credit card info in an encrypted format.
"""

from Crypto.Cipher import Blowfish
from datetime import datetime
from decimal import Decimal
from django.conf import settings
from django.db import models
from django.utils.translation import ugettext_lazy as _
from livesettings import config_value, config_choice_values, SettingNotSet
from payment.fields import PaymentChoiceCharField, CreditChoiceCharField
from satchmo_store.contact.models import Contact
from satchmo_store.shop.models import OrderPayment
import base64
import config
import keyedcache
import logging
import satchmo_utils.sslurllib

log = logging.getLogger('payment.models')
        
class PaymentOption(models.Model):
    """
    If there are multiple options - CC, Cash, COD, etc this class allows
    configuration.
    """
    description = models.CharField(_("Description"), max_length=20)
    active = models.BooleanField(_("Active"), 
        help_text=_("Should this be displayed as an option for the user?"))
    optionName = PaymentChoiceCharField(_("Option Name"), max_length=20, 
        unique=True, 
        help_text=_("The class name as defined in payment.py"))
    sortOrder = models.IntegerField(_("Sort Order"))
    
    class Meta:
        verbose_name = _("Payment Option")
        verbose_name_plural = _("Payment Options")

class CreditCardDetail(models.Model):
    """
    Stores an encrypted CC number, its information, and its
    displayable number.
    """
    orderpayment = models.ForeignKey(OrderPayment, unique=True, 
        related_name="creditcards")
    credit_type = CreditChoiceCharField(_("Credit Card Type"), max_length=16)
    display_cc = models.CharField(_("CC Number (Last 4 digits)"),
        max_length=4, )
    encrypted_cc = models.CharField(_("Encrypted Credit Card"),
        max_length=40, blank=True, null=True, editable=False)
    expire_month = models.IntegerField(_("Expiration Month"))
    expire_year = models.IntegerField(_("Expiration Year"))
    card_holder = models.CharField(_("card_holder Name"), max_length=60, blank=True)
    start_month = models.IntegerField(_("Start Month"), blank=True, null=True)
    start_year = models.IntegerField(_("Start Year"), blank=True, null=True)
    issue_num = models.CharField(blank=True, null=True, max_length=2)
    
    def storeCC(self, ccnum):
        """Take as input a valid cc, encrypt it and store the last 4 digits in a visible form"""
        self.display_cc = ccnum[-4:]
        encrypted_cc = _encrypt_code(ccnum)
        if config_value('PAYMENT', 'STORE_CREDIT_NUMBERS'):
            self.encrypted_cc = encrypted_cc
        else:
            standin = "%s%i%i%i" % (self.display_cc, self.expire_month, self.expire_year, self.orderpayment.id)
            self.encrypted_cc = _encrypt_code(standin)
            key = _encrypt_code(standin + '-card')
            keyedcache.cache_set(key, skiplog=True, length=60*60, value=encrypted_cc)
    
    def setCCV(self, ccv):
        """Put the CCV in the cache, don't save it for security/legal reasons."""
        if not self.encrypted_cc:
            raise ValueError('CreditCardDetail expecting a credit card number to be stored before storing CCV')
            
        keyedcache.cache_set(self.encrypted_cc, skiplog=True, length=60*60, value=ccv)
    
    def getCCV(self):
        try:
            ccv = keyedcache.cache_get(self.encrypted_cc)
        except keyedcache.NotCachedError:
            ccv = ""

        return ccv
    
    ccv = property(fget=getCCV, fset=setCCV)
    
    def _decryptCC(self):
        ccnum = _decrypt_code(self.encrypted_cc)
        if not config_value('PAYMENT', 'STORE_CREDIT_NUMBERS'):
            try:
                key = _encrypt_code(ccnum + '-card')
                encrypted_ccnum = keyedcache.cache_get(key)
                ccnum = _decrypt_code(encrypted_ccnum)
            except keyedcache.NotCachedError:
                ccnum = ""
        return ccnum
                
    decryptedCC = property(_decryptCC) 

    def _expireDate(self):
        return(str(self.expire_month) + "/" + str(self.expire_year))
    expirationDate = property(_expireDate)
    
    class Meta:
        verbose_name = _("Credit Card")
        verbose_name_plural = _("Credit Cards")

def _decrypt_code(code):
    """Decrypt code encrypted by _encrypt_code"""
    secret_key = settings.SECRET_KEY
    encryption_object = Blowfish.new(secret_key)
    # strip padding from decrypted credit card number
    return encryption_object.decrypt(base64.b64decode(code)).rstrip('X')

def _encrypt_code(code):
    """Quick encrypter for CC codes or code fragments"""
    secret_key = settings.SECRET_KEY
    encryption_object = Blowfish.new(secret_key)
    # block cipher length must be a multiple of 8
    padding = ''
    if (len(code) % 8) <> 0:
        padding = 'X' * (8 - (len(code) % 8))
    return base64.b64encode(encryption_object.encrypt(code + padding))



from django.db import models
from livesettings import config_value_safe, config_choice_values, SettingNotSet

def shipping_choices():
    try:
        return config_choice_values('SHIPPING','MODULES')
    except SettingNotSet:
        return ()    


class ShippingChoiceCharField(models.CharField):
    
    def __init__(self, choices="__DYNAMIC__", *args, **kwargs):
        if choices == "__DYNAMIC__":
            kwargs['choices'] = shipping_choices()
                    
        super(ShippingChoiceCharField, self).__init__(*args, **kwargs)



'''
This module provides the ``Entity`` base class, as well as its metaclass
``EntityMeta``.
'''

from py23compat import sorted

import sys
import types
import warnings

from copy import deepcopy

import sqlalchemy
from sqlalchemy import Table, Column, Integer, desc, ForeignKey, and_, \
                       ForeignKeyConstraint
from sqlalchemy.orm import MapperExtension, mapper, object_session, \
                           EXT_CONTINUE, polymorphic_union, scoped_session, \
                           ColumnProperty
from sqlalchemy.sql import ColumnCollection

import elixir
from elixir.statements import process_mutators, MUTATORS
from elixir import options
from elixir.properties import Property

DEBUG = False
try:
    from sqlalchemy.orm import EXT_PASS
    SA05orlater = False
except ImportError:
    SA05orlater = True

__doc_all__ = ['Entity', 'EntityMeta']


def session_mapper_factory(scoped_session):
    def session_mapper(cls, *args, **kwargs):
        if kwargs.pop('save_on_init', True):
            old_init = cls.__init__
            def __init__(self, *args, **kwargs):
                old_init(self, *args, **kwargs)
                scoped_session.add(self)
            cls.__init__ = __init__
        cls.query = scoped_session.query_property()
        return mapper(cls, *args, **kwargs)
    return session_mapper


class EntityDescriptor(object):
    '''
    EntityDescriptor describes fields and options needed for table creation.
    '''

    def __init__(self, entity):
        self.entity = entity
        self.parent = None

        bases = []
        for base in entity.__bases__:
            if isinstance(base, EntityMeta):
                if is_entity(base) and not is_abstract_entity(base):
                    if self.parent:
                        raise Exception(
                            '%s entity inherits from several entities, '
                            'and this is not supported.'
                            % self.entity.__name__)
                    else:
                        self.parent = base
                        bases.extend(base._descriptor.bases)
                        self.parent._descriptor.children.append(entity)
                else:
                    bases.append(base)
        self.bases = bases
        if not is_entity(entity) or is_abstract_entity(entity):
            return

        # entity.__module__ is not always reliable (eg in mod_python)
        self.module = sys.modules.get(entity.__module__)

        self.builders = []

        #XXX: use entity.__subclasses__ ?
        self.children = []

        # used for multi-table inheritance
        self.join_condition = None
        self.has_pk = False
        self._pk_col_done = False

        # columns and constraints waiting for a table to exist
        self._columns = ColumnCollection()
        self.constraints = []

        # properties (it is only useful for checking dupe properties at the
        # moment, and when adding properties before the mapper is created,
        # which shouldn't happen).
        self.properties = {}

        #
        self.relationships = []

        # set default value for options
        self.table_args = []

        # base class(es) options_defaults
        options_defaults = self.options_defaults()

        complete_defaults = options.options_defaults.copy()
        complete_defaults.update({
            'metadata': elixir.metadata,
            'session': elixir.session,
            'collection': elixir.entities
        })

        # set default value for other options
        for key in options.valid_options:
            value = options_defaults.get(key, complete_defaults[key])
            if isinstance(value, dict):
                value = value.copy()
            setattr(self, key, value)

        # override options with module-level defaults defined
        for key in ('metadata', 'session', 'collection'):
            attr = '__%s__' % key
            if hasattr(self.module, attr):
                setattr(self, key, getattr(self.module, attr))

    def options_defaults(self):
        base_defaults = {}
        for base in self.bases:
            base_defaults.update(base._descriptor.options_defaults())
        base_defaults.update(getattr(self.entity, 'options_defaults', {}))
        return base_defaults

    def setup_options(self):
        '''
        Setup any values that might depend on the "using_options" class
        mutator. For example, the tablename or the metadata.
        '''
        elixir.metadatas.add(self.metadata)
        if self.collection is not None:
            self.collection.append(self.entity)

        entity = self.entity
        if self.parent:
            if self.inheritance == 'single':
                self.tablename = self.parent._descriptor.tablename

        if not self.tablename:
            if self.shortnames:
                self.tablename = entity.__name__.lower()
            else:
                modulename = entity.__module__.replace('.', '_')
                tablename = "%s_%s" % (modulename, entity.__name__)
                self.tablename = tablename.lower()
        elif hasattr(self.tablename, '__call__'):
            self.tablename = self.tablename(entity)

        if not self.identity:
            if 'polymorphic_identity' in self.mapper_options:
                self.identity = self.mapper_options['polymorphic_identity']
            else:
                #TODO: include module name (We could have b.Account inherit
                # from a.Account)
                self.identity = entity.__name__.lower()
        elif 'polymorphic_identity' in self.mapper_options:
            raise Exception('You cannot use the "identity" option and the '
                            'polymorphic_identity mapper option at the same '
                            'time.')
        elif hasattr(self.identity, '__call__'):
            self.identity = self.identity(entity)

        if self.polymorphic:
            if not isinstance(self.polymorphic, basestring):
                self.polymorphic = options.DEFAULT_POLYMORPHIC_COL_NAME

    #---------------------
    # setup phase methods

    def setup_autoload_table(self):
        self.setup_table(True)

    def create_pk_cols(self):
        """
        Create primary_key columns. That is, call the 'create_pk_cols'
        builders then add a primary key to the table if it hasn't already got
        one and needs one.

        This method is "semi-recursive" in some cases: it calls the
        create_keys method on ManyToOne relationships and those in turn call
        create_pk_cols on their target. It shouldn't be possible to have an
        infinite loop since a loop of primary_keys is not a valid situation.
        """
        if self._pk_col_done:
            return

        self.call_builders('create_pk_cols')

        if not self.autoload:
            if self.parent:
                if self.inheritance == 'multi':
                    # Add columns with foreign keys to the parent's primary
                    # key columns
                    parent_desc = self.parent._descriptor
                    tablename = parent_desc.table_fullname
                    join_clauses = []
                    for pk_col in parent_desc.primary_keys:
                        colname = options.MULTIINHERITANCECOL_NAMEFORMAT % \
                                  {'entity': self.parent.__name__.lower(),
                                   'key': pk_col.key}

                        # It seems like SA ForeignKey is not happy being given
                        # a real column object when said column is not yet
                        # attached to a table
                        pk_col_name = "%s.%s" % (tablename, pk_col.key)
                        fk = ForeignKey(pk_col_name, ondelete='cascade')
                        col = Column(colname, pk_col.type, fk,
                                     primary_key=True)
                        self.add_column(col)
                        join_clauses.append(col == pk_col)
                    self.join_condition = and_(*join_clauses)
                elif self.inheritance == 'concrete':
                    # Copy primary key columns from the parent.
                    for col in self.parent._descriptor.columns:
                        if col.primary_key:
                            self.add_column(col.copy())
            elif not self.has_pk and self.auto_primarykey:
                if isinstance(self.auto_primarykey, basestring):
                    colname = self.auto_primarykey
                else:
                    colname = options.DEFAULT_AUTO_PRIMARYKEY_NAME

                self.add_column(
                    Column(colname, options.DEFAULT_AUTO_PRIMARYKEY_TYPE,
                           primary_key=True))
        self._pk_col_done = True

    def setup_relkeys(self):
        self.call_builders('create_non_pk_cols')

    def before_table(self):
        self.call_builders('before_table')

    def setup_table(self, only_autoloaded=False):
        '''
        Create a SQLAlchemy table-object with all columns that have been
        defined up to this point.
        '''
        if self.entity.table is not None:
            return

        if self.autoload != only_autoloaded:
            return

        kwargs = self.table_options
        if self.autoload:
            args = self.table_args
            kwargs['autoload'] = True
        else:
            if self.parent:
                if self.inheritance == 'single':
                    # we know the parent is setup before the child
                    self.entity.table = self.parent.table

                    # re-add the entity columns to the parent entity so that
                    # they are added to the parent's table (whether the
                    # parent's table is already setup or not).
                    for col in self._columns:
                        self.parent._descriptor.add_column(col)
                    for constraint in self.constraints:
                        self.parent._descriptor.add_constraint(constraint)
                    return
                elif self.inheritance == 'concrete':
                    #TODO: we should also copy columns from the parent table
                    # if the parent is a base (abstract?) entity (whatever the
                    # inheritance type -> elif will need to be changed)

                    # Copy all non-primary key columns from parent table
                    # (primary key columns have already been copied earlier).
                    for col in self.parent._descriptor.columns:
                        if not col.primary_key:
                            self.add_column(col.copy())

                    for con in self.parent._descriptor.constraints:
                        self.add_constraint(
                            ForeignKeyConstraint(
                                [e.parent.key for e in con.elements],
                                [e._get_colspec() for e in con.elements],
                                name=con.name, #TODO: modify it
                                onupdate=con.onupdate, ondelete=con.ondelete,
                                use_alter=con.use_alter))

            if self.polymorphic and \
               self.inheritance in ('single', 'multi') and \
               self.children and not self.parent:
                self.add_column(Column(self.polymorphic,
                                       options.POLYMORPHIC_COL_TYPE))

            if self.version_id_col:
                if not isinstance(self.version_id_col, basestring):
                    self.version_id_col = options.DEFAULT_VERSION_ID_COL_NAME
                self.add_column(Column(self.version_id_col, Integer))

            args = list(self.columns) + self.constraints + self.table_args
        self.entity.table = Table(self.tablename, self.metadata,
                                  *args, **kwargs)
        if DEBUG:
            print self.entity.table.repr2()

    def setup_reltables(self):
        self.call_builders('create_tables')

    def after_table(self):
        self.call_builders('after_table')

    def setup_events(self):
        def make_proxy_method(methods):
            def proxy_method(self, mapper, connection, instance):
                for func in methods:
                    ret = func(instance)
                    # I couldn't commit myself to force people to
                    # systematicaly return EXT_CONTINUE in all their event
                    # methods.
                    # But not doing that diverge to how SQLAlchemy works.
                    # I should try to convince Mike to do EXT_CONTINUE by
                    # default, and stop processing as the special case.
#                    if ret != EXT_CONTINUE:
                    if ret is not None and ret != EXT_CONTINUE:
                        return ret
                return EXT_CONTINUE
            return proxy_method

        # create a list of callbacks for each event
        methods = {}

        all_methods = getmembers(self.entity,
                                 lambda a: isinstance(a, types.MethodType))

        for name, method in all_methods:
            for event in getattr(method, '_elixir_events', []):
                event_methods = methods.setdefault(event, [])
                event_methods.append(method)

        if not methods:
            return

        # transform that list into methods themselves
        for event in methods:
            methods[event] = make_proxy_method(methods[event])

        # create a custom mapper extension class, tailored to our entity
        ext = type('EventMapperExtension', (MapperExtension,), methods)()

        # then, make sure that the entity's mapper has our mapper extension
        self.add_mapper_extension(ext)

    def before_mapper(self):
        self.call_builders('before_mapper')

    def _get_children(self):
        children = self.children[:]
        for child in self.children:
            children.extend(child._descriptor._get_children())
        return children

    def translate_order_by(self, order_by):
        if isinstance(order_by, basestring):
            order_by = [order_by]

        order = []
        for colname in order_by:
            col = self.get_column(colname.strip('-'))
            if colname.startswith('-'):
                col = desc(col)
            order.append(col)
        return order

    def setup_mapper(self):
        '''
        Initializes and assign a mapper to the entity.
        At this point the mapper will usually have no property as they are
        added later.
        '''
        if self.entity.mapper:
            return

        # for now we don't support the "abstract" parent class in a concrete
        # inheritance scenario as demonstrated in
        # sqlalchemy/test/orm/inheritance/concrete.py
        # this should be added along other
        kwargs = {}
        if self.order_by:
            kwargs['order_by'] = self.translate_order_by(self.order_by)

        if self.version_id_col:
            kwargs['version_id_col'] = self.get_column(self.version_id_col)

        if self.inheritance in ('single', 'concrete', 'multi'):
            if self.parent and \
               (self.inheritance != 'concrete' or self.polymorphic):
                # non-polymorphic concrete doesn't need this
                kwargs['inherits'] = self.parent.mapper

            if self.inheritance == 'multi' and self.parent:
                kwargs['inherit_condition'] = self.join_condition

            if self.polymorphic:
                if self.children:
                    if self.inheritance == 'concrete':
                        keys = [(self.identity, self.entity.table)]
                        keys.extend([(child._descriptor.identity, child.table)
                                     for child in self._get_children()])
                        # Having the same alias name for an entity and one of
                        # its child (which is a parent itself) shouldn't cause
                        # any problem because the join shouldn't be used at
                        # the same time. But in reality, some versions of SA
                        # do misbehave on this. Since it doesn't hurt to have
                        # different names anyway, here they go.
                        pjoin = polymorphic_union(
                                    dict(keys), self.polymorphic,
                                    'pjoin_%s' % self.identity)

                        kwargs['with_polymorphic'] = ('*', pjoin)
                        kwargs['polymorphic_on'] = \
                            getattr(pjoin.c, self.polymorphic)
                    elif not self.parent:
                        kwargs['polymorphic_on'] = \
                            self.get_column(self.polymorphic)

                if self.children or self.parent:
                    kwargs['polymorphic_identity'] = self.identity

                if self.parent and self.inheritance == 'concrete':
                    kwargs['concrete'] = True

        if self.parent and self.inheritance == 'single':
            args = []
        else:
            args = [self.entity.table]

        # let user-defined kwargs override Elixir-generated ones, though that's
        # not very usefull since most of them expect Column instances.
        kwargs.update(self.mapper_options)

        #TODO: document this!
        if 'primary_key' in kwargs:
            cols = self.entity.table.c
            kwargs['primary_key'] = [getattr(cols, colname) for
                colname in kwargs['primary_key']]

        # do the mapping
        if self.session is None:
            self.entity.mapper = mapper(self.entity, *args, **kwargs)
        elif isinstance(self.session, scoped_session):
            session_mapper = session_mapper_factory(self.session)
            self.entity.mapper = session_mapper(self.entity, *args, **kwargs)
        else:
            raise Exception("Failed to map entity '%s' with its table or "
                            "selectable. You can only bind an Entity to a "
                            "ScopedSession object or None for manual session "
                            "management."
                            % self.entity.__name__)

    def after_mapper(self):
        self.call_builders('after_mapper')

    def setup_properties(self):
        self.call_builders('create_properties')

    def finalize(self):
        self.call_builders('finalize')
        self.entity._setup_done = True

    #----------------
    # helper methods

    def call_builders(self, what):
        for builder in self.builders:
            if hasattr(builder, what):
                getattr(builder, what)()

    def add_column(self, col, check_duplicate=None):
        '''when check_duplicate is None, the value of the allowcoloverride
        option of the entity is used.
        '''
        if check_duplicate is None:
            check_duplicate = not self.allowcoloverride

        if col.key in self._columns:
            if check_duplicate:
                raise Exception("Column '%s' already exist in '%s' ! " %
                                (col.key, self.entity.__name__))
            else:
                del self._columns[col.key]
        self._columns.add(col)

        if col.primary_key:
            self.has_pk = True

        # Autosetup triggers shouldn't be active anymore at this point, so we
        # can theoretically access the entity's table safely. But the problem
        # is that if, for some reason, the trigger removal phase didn't
        # happen, we'll get an infinite loop. So we just make sure we don't
        # get one in any case.
        table = type.__getattribute__(self.entity, 'table')
        if table is not None:
            if check_duplicate and col.key in table.columns.keys():
                raise Exception("Column '%s' already exist in table '%s' ! " %
                                (col.key, table.name))
            table.append_column(col)
            if DEBUG:
                print "table.append_column(%s)" % col

    def add_constraint(self, constraint):
        self.constraints.append(constraint)

        table = self.entity.table
        if table is not None:
            table.append_constraint(constraint)

    def add_property(self, name, property, check_duplicate=True):
        if check_duplicate and name in self.properties:
            raise Exception("property '%s' already exist in '%s' ! " %
                            (name, self.entity.__name__))
        self.properties[name] = property

#FIXME: something like this is needed to propagate the relationships from
# parent entities to their children in a concrete inheritance scenario. But
# this doesn't work because of the backref matching code. In most case
# (test_concrete.py) it doesn't even happen at all.
#        if self.children and self.inheritance == 'concrete':
#            for child in self.children:
#                child._descriptor.add_property(name, property)

        mapper = self.entity.mapper
        if mapper:
            mapper.add_property(name, property)
            if DEBUG:
                print "mapper.add_property('%s', %s)" % (name, repr(property))

    def add_mapper_extension(self, extension):
        extensions = self.mapper_options.get('extension', [])
        if not isinstance(extensions, list):
            extensions = [extensions]
        extensions.append(extension)
        self.mapper_options['extension'] = extensions

    def get_column(self, key, check_missing=True):
        #TODO: this needs to work whether the table is already setup or not
        #TODO: support SA table/autoloaded entity
        try:
            return self.columns[key]
        except KeyError:
            if check_missing:
                raise Exception("No column named '%s' found in the table of "
                                "the '%s' entity!"
                                % (key, self.entity.__name__))

    def get_inverse_relation(self, rel, check_reverse=True):
        '''
        Return the inverse relation of rel, if any, None otherwise.
        '''

        matching_rel = None
        for other_rel in self.relationships:
            if rel.is_inverse(other_rel):
                if matching_rel is None:
                    matching_rel = other_rel
                else:
                    raise Exception(
                            "Several relations match as inverse of the '%s' "
                            "relation in entity '%s'. You should specify "
                            "inverse relations manually by using the inverse "
                            "keyword."
                            % (rel.name, rel.entity.__name__))
        # When a matching inverse is found, we check that it has only
        # one relation matching as its own inverse. We don't need the result
        # of the method though. But we do need to be careful not to start an
        # infinite recursive loop.
        if matching_rel and check_reverse:
            rel.entity._descriptor.get_inverse_relation(matching_rel, False)

        return matching_rel

    def find_relationship(self, name):
        for rel in self.relationships:
            if rel.name == name:
                return rel
        if self.parent:
            return self.parent._descriptor.find_relationship(name)
        else:
            return None

    #------------------------
    # some useful properties

    def table_fullname(self):
        '''
        Complete name of the table for the related entity.
        Includes the schema name if there is one specified.
        '''
        schema = self.table_options.get('schema', None)
        if schema is not None:
            return "%s.%s" % (schema, self.tablename)
        else:
            return self.tablename
    table_fullname = property(table_fullname)

    def columns(self):
        if self.entity.table is not None:
            return self.entity.table.columns
        else:
            #FIXME: depending on the type of inheritance, we should also
            # return the parent entity's columns (for example for order_by
            # using a column defined in the parent.
            return self._columns
    columns = property(columns)

    def primary_keys(self):
        """
        Returns the list of primary key columns of the entity.

        This property isn't valid before the "create_pk_cols" phase.
        """
        if self.autoload:
            return [col for col in self.entity.table.primary_key.columns]
        else:
            if self.parent and self.inheritance == 'single':
                return self.parent._descriptor.primary_keys
            else:
                return [col for col in self.columns if col.primary_key]
    primary_keys = property(primary_keys)

    def table(self):
        if self.entity.table is not None:
            return self.entity.table
        else:
            return FakeTable(self)
    table = property(table)

    def primary_key_properties(self):
        """
        Returns the list of (mapper) properties corresponding to the primary
        key columns of the table of the entity.

        This property caches its value, so it shouldn't be called before the
        entity is fully set up.
        """
        if not hasattr(self, '_pk_props'):
            col_to_prop = {}
            mapper = self.entity.mapper
            for prop in mapper.iterate_properties:
                if isinstance(prop, ColumnProperty):
                    for col in prop.columns:
                        for col in col.proxy_set:
                            col_to_prop[col] = prop
            pk_cols = [c for c in mapper.mapped_table.c if c.primary_key]
            self._pk_props = [col_to_prop[c] for c in pk_cols]
        return self._pk_props
    primary_key_properties = property(primary_key_properties)

class FakePK(object):
    def __init__(self, descriptor):
        self.descriptor = descriptor

    def columns(self):
        return self.descriptor.primary_keys
    columns = property(columns)

class FakeTable(object):
    def __init__(self, descriptor):
        self.descriptor = descriptor
        self.primary_key = FakePK(descriptor)

    def columns(self):
        return self.descriptor.columns
    columns = property(columns)

    def fullname(self):
        '''
        Complete name of the table for the related entity.
        Includes the schema name if there is one specified.
        '''
        schema = self.descriptor.table_options.get('schema', None)
        if schema is not None:
            return "%s.%s" % (schema, self.descriptor.tablename)
        else:
            return self.descriptor.tablename
    fullname = property(fullname)


class TriggerProxy(object):
    """
    A class that serves as a "trigger" ; accessing its attributes runs
    the setup_all function.

    Note that the `setup_all` is called on each access of the attribute.
    """

    def __init__(self, class_, attrname):
        self.class_ = class_
        self.attrname = attrname

    def __getattr__(self, name):
        elixir.setup_all()
        #FIXME: it's possible to get an infinite loop here if setup_all doesn't
        #remove the triggers for this entity. This can happen if the entity is
        #not in the `entities` list for some reason.
        proxied_attr = getattr(self.class_, self.attrname)
        return getattr(proxied_attr, name)

    def __repr__(self):
        proxied_attr = getattr(self.class_, self.attrname)
        return "<TriggerProxy (%s)>" % (self.class_.__name__)


class TriggerAttribute(object):

    def __init__(self, attrname):
        self.attrname = attrname

    def __get__(self, instance, owner):
        #FIXME: it's possible to get an infinite loop here if setup_all doesn't
        #remove the triggers for this entity. This can happen if the entity is
        #not in the `entities` list for some reason.
        elixir.setup_all()
        return getattr(owner, self.attrname)

def is_entity(cls):
    """
    Scan the bases classes of `cls` to see if any is an instance of
    EntityMeta. If we don't find any, it means it is either an unrelated class
    or an entity base class (like the 'Entity' class).
    """
    for base in cls.__bases__:
        if isinstance(base, EntityMeta):
            return True
    return False


# Note that we don't use inspect.getmembers because of
# http://bugs.python.org/issue1785
# See also http://elixir.ematia.de/trac/changeset/262
def getmembers(object, predicate=None):
    base_props = []
    for key in dir(object):
        try:
            value = getattr(object, key)
        except AttributeError:
            continue
        if not predicate or predicate(value):
            base_props.append((key, value))
    return base_props

def is_abstract_entity(dict_or_cls):
    if not isinstance(dict_or_cls, dict):
        dict_or_cls = dict_or_cls.__dict__
    for mutator, args, kwargs in dict_or_cls.get(MUTATORS, []):
        if 'abstract' in kwargs:
            return kwargs['abstract']

    return False

def instrument_class(cls):
    """
    Instrument a class as an Entity. This is usually done automatically through
    the EntityMeta metaclass.
    """
    # Create the entity descriptor
    desc = cls._descriptor = EntityDescriptor(cls)

    # Process mutators
    # We *do* want mutators to be processed for base/abstract classes
    # (so that statements like using_options_defaults work).
    process_mutators(cls)

    # We do not want to do any more processing for base/abstract classes
    # (Entity et al.).
    if not is_entity(cls) or is_abstract_entity(cls):
        return

    cls.table = None
    cls.mapper = None

    # Copy the properties ('Property' instances) of the entity base class(es).
    # We use getmembers (instead of __dict__) so that we also get the
    # properties from the parents of the base class if any.
    base_props = []
    for base in cls.__bases__:
        if isinstance(base, EntityMeta) and \
           (not is_entity(base) or is_abstract_entity(base)):
            base_props += [(name, deepcopy(attr)) for name, attr in
                           getmembers(base, lambda a: isinstance(a, Property))]

    # Process attributes (using the assignment syntax), looking for
    # 'Property' instances and attaching them to this entity.
    properties = [(name, attr) for name, attr in cls.__dict__.iteritems()
                               if isinstance(attr, Property)]
    sorted_props = sorted(base_props + properties,
                          key=lambda i: i[1]._counter)
    for name, prop in sorted_props:
        prop.attach(cls, name)

    # setup misc options here (like tablename etc.)
    desc.setup_options()

    # create trigger proxies
    # TODO: support entity_name... It makes sense only for autoloaded
    # tables for now, and would make more sense if we support "external"
    # tables
    if desc.autosetup:
        _install_autosetup_triggers(cls)


class EntityMeta(type):
    """
    Entity meta class.
    You should only use it directly if you want to define your own base class
    for your entities (ie you don't want to use the provided 'Entity' class).
    """

    def __init__(cls, name, bases, dict_):
        instrument_class(cls)

    def __call__(cls, *args, **kwargs):
        if cls._descriptor.autosetup and not hasattr(cls, '_setup_done'):
            elixir.setup_all()
        return type.__call__(cls, *args, **kwargs)

    def __setattr__(cls, key, value):
        if isinstance(value, Property):
            if hasattr(cls, '_setup_done'):
                raise Exception('Cannot set attribute on a class after '
                                'setup_all')
            else:
                value.attach(cls, key)
        else:
            type.__setattr__(cls, key, value)


def _install_autosetup_triggers(cls, entity_name=None):
    #TODO: move as much as possible of those "_private" values to the
    # descriptor, so that we don't mess the initial class.
    warnings.warn("The 'autosetup' option on entities is deprecated. "
        "Please call setup_all() manually after all your entities have been "
        "declared.", DeprecationWarning, stacklevel=4)
    tablename = cls._descriptor.tablename
    schema = cls._descriptor.table_options.get('schema', None)
    cls._table_key = sqlalchemy.schema._get_table_key(tablename, schema)

    table_proxy = TriggerProxy(cls, 'table')

    md = cls._descriptor.metadata
    md.tables[cls._table_key] = table_proxy

    # We need to monkeypatch the metadata's table iterator method because
    # otherwise it doesn't work if the setup is triggered by the
    # metadata.create_all().
    # This is because ManyToMany relationships add tables AFTER the list
    # of tables that are going to be created is "computed"
    # (metadata.tables.values()).
    # see:
    # - table_iterator method in MetaData class in sqlalchemy/schema.py
    # - visit_metadata method in sqlalchemy/ansisql.py
    if SA05orlater:
        warnings.warn(
            "The automatic setup via metadata.create_all() through "
            "the autosetup option doesn't work with SQLAlchemy 0.5 and later!")
    else:
        # SA 0.6 does not use table_iterator anymore (it was already deprecated
        # since SA 0.5.0)
        original_table_iterator = md.table_iterator
        if not hasattr(original_table_iterator,
                       '_non_elixir_patched_iterator'):
            def table_iterator(*args, **kwargs):
                elixir.setup_all()
                return original_table_iterator(*args, **kwargs)
            table_iterator.__doc__ = original_table_iterator.__doc__
            table_iterator._non_elixir_patched_iterator = \
                original_table_iterator
            md.table_iterator = table_iterator

    #TODO: we might want to add all columns that will be available as
    #attributes on the class itself (in SA 0.4+). This is a pretty
    #rare usecase, as people will normally hit the query attribute before the
    #column attributes, but I've seen people hitting this problem...
    for name in ('c', 'table', 'mapper', 'query'):
        setattr(cls, name, TriggerAttribute(name))

    cls._has_triggers = True


def _cleanup_autosetup_triggers(cls):
    if not hasattr(cls, '_has_triggers'):
        return

    for name in ('table', 'mapper'):
        setattr(cls, name, None)

    for name in ('c', 'query'):
        delattr(cls, name)

    desc = cls._descriptor
    md = desc.metadata

    # the fake table could have already been removed (namely in a
    # single table inheritance scenario)
    md.tables.pop(cls._table_key, None)

    # restore original table iterator if not done already
    if not SA05orlater:
        if hasattr(md.table_iterator, '_non_elixir_patched_iterator'):
            md.table_iterator = \
                md.table_iterator._non_elixir_patched_iterator

    del cls._has_triggers


def setup_entities(entities):
    '''Setup all entities in the list passed as argument'''

    for entity in entities:
        # delete all Elixir properties so that it doesn't interfere with
        # SQLAlchemy. At this point they should have be converted to
        # builders.
        for name, attr in entity.__dict__.items():
            if isinstance(attr, Property):
                delattr(entity, name)

        if entity._descriptor.autosetup:
            _cleanup_autosetup_triggers(entity)

    for method_name in (
            'setup_autoload_table', 'create_pk_cols', 'setup_relkeys',
            'before_table', 'setup_table', 'setup_reltables', 'after_table',
            'setup_events',
            'before_mapper', 'setup_mapper', 'after_mapper',
            'setup_properties',
            'finalize'):
#        if DEBUG:
#            print "=" * 40
#            print method_name
#            print "=" * 40
        for entity in entities:
#            print entity.__name__, "...",
            if hasattr(entity, '_setup_done'):
#                print "already done"
                continue
            method = getattr(entity._descriptor, method_name)
            method()
#            print "ok"


def cleanup_entities(entities):
    """
    Try to revert back the list of entities passed as argument to the state
    they had just before their setup phase. It will not work entirely for
    autosetup entities as we need to remove the autosetup triggers.

    As of now, this function is *not* functional in that it doesn't revert to
    the exact same state the entities were before setup. For example, the
    properties do not work yet as those would need to be regenerated (since the
    columns they are based on are regenerated too -- and as such the
    corresponding joins are not correct) but this doesn't happen because of
    the way relationship setup is designed to be called only once (especially
    the backref stuff in create_properties).
    """
    for entity in entities:
        desc = entity._descriptor
        if desc.autosetup:
            _cleanup_autosetup_triggers(entity)

        if hasattr(entity, '_setup_done'):
            del entity._setup_done

        entity.table = None
        entity.mapper = None

        desc._pk_col_done = False
        desc.has_pk = False
        desc._columns = ColumnCollection()
        desc.constraints = []
        desc.properties = {}

class EntityBase(object):
    """
    This class holds all methods of the "Entity" base class, but does not act
    as a base class itself (it does not use the EntityMeta metaclass), but
    rather as a parent class for Entity. This is meant so that people who want
    to provide their own base class but don't want to loose or copy-paste all
    the methods of Entity can do so by inheriting from EntityBase:

    .. sourcecode:: python

        class MyBase(EntityBase):
            __metaclass__ = EntityMeta

            def myCustomMethod(self):
                # do something great
    """

    def __init__(self, **kwargs):
        self.set(**kwargs)

    def set(self, **kwargs):
        for key, value in kwargs.iteritems():
            setattr(self, key, value)

    def update_or_create(cls, data, surrogate=True):
        pk_props = cls._descriptor.primary_key_properties

        # if all pk are present and not None
        if not [1 for p in pk_props if data.get(p.key) is None]:
            pk_tuple = tuple([data[prop.key] for prop in pk_props])
            record = cls.query.get(pk_tuple)
            if record is None:
                if surrogate:
                    raise Exception("cannot create surrogate with pk")
                else:
                    record = cls()
        else:
            if surrogate:
                record = cls()
            else:
                raise Exception("cannot create non surrogate without pk")
        record.from_dict(data)
        return record
    update_or_create = classmethod(update_or_create)

    def from_dict(self, data):
        """
        Update a mapped class with data from a JSON-style nested dict/list
        structure.
        """
        # surrogate can be guessed from autoincrement/sequence but I guess
        # that's not 100% reliable, so we'll need an override

        mapper = sqlalchemy.orm.object_mapper(self)

        for key, value in data.iteritems():
            if isinstance(value, dict):
                dbvalue = getattr(self, key)
                rel_class = mapper.get_property(key).mapper.class_
                pk_props = rel_class._descriptor.primary_key_properties

                # If the data doesn't contain any pk, and the relationship
                # already has a value, update that record.
                if not [1 for p in pk_props if p.key in data] and \
                   dbvalue is not None:
                    dbvalue.from_dict(value)
                else:
                    record = rel_class.update_or_create(value)
                    setattr(self, key, record)
            elif isinstance(value, list) and \
                 value and isinstance(value[0], dict):

                rel_class = mapper.get_property(key).mapper.class_
                new_attr_value = []
                for row in value:
                    if not isinstance(row, dict):
                        raise Exception(
                                'Cannot send mixed (dict/non dict) data '
                                'to list relationships in from_dict data.')
                    record = rel_class.update_or_create(row)
                    new_attr_value.append(record)
                setattr(self, key, new_attr_value)
            else:
                setattr(self, key, value)

    def to_dict(self, deep={}, exclude=[]):
        """Generate a JSON-style nested dict/list structure from an object."""
        col_prop_names = [p.key for p in self.mapper.iterate_properties \
                                      if isinstance(p, ColumnProperty)]
        data = dict([(name, getattr(self, name))
                     for name in col_prop_names if name not in exclude])
        for rname, rdeep in deep.iteritems():
            dbdata = getattr(self, rname)
            #FIXME: use attribute names (ie coltoprop) instead of column names
            fks = self.mapper.get_property(rname).remote_side
            exclude = [c.name for c in fks]
            if dbdata is None:
                data[rname] = None
            elif isinstance(dbdata, list):
                data[rname] = [o.to_dict(rdeep, exclude) for o in dbdata]
            else:
                data[rname] = dbdata.to_dict(rdeep, exclude)
        return data

    # session methods
    def flush(self, *args, **kwargs):
        return object_session(self).flush([self], *args, **kwargs)

    def delete(self, *args, **kwargs):
        return object_session(self).delete(self, *args, **kwargs)

    def expire(self, *args, **kwargs):
        return object_session(self).expire(self, *args, **kwargs)

    def refresh(self, *args, **kwargs):
        return object_session(self).refresh(self, *args, **kwargs)

    def expunge(self, *args, **kwargs):
        return object_session(self).expunge(self, *args, **kwargs)

    # This bunch of session methods, along with all the query methods below
    # only make sense when using a global/scoped/contextual session.
    def _global_session(self):
        return self._descriptor.session.registry()
    _global_session = property(_global_session)

    def merge(self, *args, **kwargs):
        return self._global_session.merge(self, *args, **kwargs)

    def save(self, *args, **kwargs):
        return self._global_session.save(self, *args, **kwargs)

    def update(self, *args, **kwargs):
        return self._global_session.update(self, *args, **kwargs)

    # only exist in SA < 0.5
    # IMO, the replacement (session.add) doesn't sound good enough to be added
    # here. For example: "o = Order(); o.add()" is not very telling. It's
    # better to leave it as "session.add(o)"
    def save_or_update(self, *args, **kwargs):
        return self._global_session.save_or_update(self, *args, **kwargs)

    # query methods
    def get_by(cls, *args, **kwargs):
        """
        Returns the first instance of this class matching the given criteria.
        This is equivalent to:
        session.query(MyClass).filter_by(...).first()
        """
        return cls.query.filter_by(*args, **kwargs).first()
    get_by = classmethod(get_by)

    def get(cls, *args, **kwargs):
        """
        Return the instance of this class based on the given identifier,
        or None if not found. This is equivalent to:
        session.query(MyClass).get(...)
        """
        return cls.query.get(*args, **kwargs)
    get = classmethod(get)


class Entity(EntityBase):
    '''
    The base class for all entities

    All Elixir model objects should inherit from this class. Statements can
    appear within the body of the definition of an entity to define its
    fields, relationships, and other options.

    Here is an example:

    .. sourcecode:: python

        class Person(Entity):
            name = Field(Unicode(128))
            birthdate = Field(DateTime, default=datetime.now)

    Please note, that if you don't specify any primary keys, Elixir will
    automatically create one called ``id``.

    For further information, please refer to the provided examples or
    tutorial.
    '''
    __metaclass__ = EntityMeta





from django.conf import settings
from django.core import urlresolvers
from django.test import TestCase
from product.models import Product
import keyedcache

domain = 'http://example.com'

class GoogleBaseTest(TestCase):
    """Test Google Base feed."""

    fixtures = ['l10n-data.yaml','sample-store-data.yaml', 'products.yaml', 'test-config.yaml']

    def tearDown(self):
        keyedcache.cache_delete

    def test_feed(self):
        url = urlresolvers.reverse('satchmo_atom_feed')
        response = self.client.get(url)
        self.assertContains(response,
            "<title>Robots Attack! (Hard cover)</title>",
            count=1, status_code=200)
        product = Product.objects.get(slug='robot-attack-hard')
        producturl = product.get_absolute_url()
        self.assertContains(response,
            "<link href=\"%s%s\" />" % (domain, producturl), count=1, status_code=200)



import os
import requests
import threading
import time

from Queue import Queue
from test_support import unittest

test_queue = Queue()

def worker(ident, url, queue, eventFlag):
    #eventFlag.wait()
    start = time.clock()
    r = requests.get(url)
    end = time.clock()
    queue.put((ident, r.status_code, end - start))
    eventFlag.set()
    #queue.task_done()

class HTTPBenchmarkTestCase(unittest.TestCase):
    
    url = 'http://localhost/benchmark/'
    threads = 5

    def setUp(self):
        self.eventFlag = threading.Event()

    def tearDown(self):
        pass

    def test_benchmark_concurrency(self):

        for i in range(self.threads):
            t = threading.Thread(target=worker, args=(i, self.url, test_queue, self.eventFlag))
            #t.daemon = True
            t.start()
            #self.eventFlag.clear()
            obj = test_queue.get()
            print obj
        #test_queue.join()

    




"""Modifies the calculations of prices to count all items in the cart when figuring quantity discounts.

To activate, simply import in your settings file.
"""

from satchmo_store.shop.signals import satchmo_cartitem_price_query
import logging
log = logging.getLogger('satchmo_ext.contrib.cartqty_discounts')

def lineitem_cartqty_price(cartitem=None, **kwargs):
    cart = cartitem.cart
    qty = cart.numItems
    oldprice = cartitem.qty_price
    newprice = cartitem.get_qty_price(qty)
    if oldprice != newprice:
        cartitem.qty_price = newprice
        log.debug("updated price for item based on cart qty=%i new price on %s = %0.2f", 
            qty, cartitem.product.slug, cartitem.qty_price)

satchmo_cartitem_price_query.connect(lineitem_cartqty_price, sender=None)
log.debug('registered lineitem_cartqty_price')


from django.contrib.comments.models import Comment
from django.contrib.comments.signals import comment_will_be_posted, comment_was_posted
from listeners import *

comment_was_posted.connect(save_rating, sender=Comment)
comment_was_posted.connect(one_rating_per_product, sender=Comment)
comment_was_posted.connect(check_with_akismet, sender=Comment)



# Copyright (C) 2005-2006 by Johan Dahlin
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

# properties.py - More black magic than Dumbledore can handle

import sys

import gobject
import gtk
from kiwi.environ import environ
from kiwi.component import implements

from gazpacho.choice import enum_to_string, flags_to_string
from gazpacho.interfaces import IReferencable
from gazpacho.command import Command
from gazpacho.i18n import _

UNICHAR_PROPERTIES = ('invisible-char', )

class PropertyError(Exception):
    pass

class UnsupportedProperty(PropertyError):
    pass

class PropertySetError(PropertyError):
    """
    This can only be raised inside a set handler of a PropType subclass.
    It is raised when an invalid value set, for example when the position
    of a cell in a table is outside of the table borders.
    """

class PropertyCustomEditor(object):
    """
    Base class for editors for custom properties

    The create method is called the first time a widget with this property is
    shown in the editor. Then, every time an instance of this type of widget is
    selected the 'update' method is called with the proper information to
    manipulate the widget.

    Because of that this should not store any information about the state of
    the widget that is being edited.
    """

    # if wide_editor is True the editor will expand both columns of the
    # editor page
    wide_editor = False

    def get_editor_widget(self):
        "Return the gtk widget that implements this editor"

    def update(self, context, widget, proxy):
        """
        Called when a widget with this property is selected

        @param    context: gazpacho context to get the current app and project
        @param widget: the widget we are editing
        @param      proxy: an object that should be used to update the widget.
                           This is what allows the undo/redo mechanism actually
                           works
        """

class PropStorage:
    """A PropStorage is a bag where we keep PropTypes

    It is used in PropRegistry to store the regular properties and the child
    properties.
    """
    child = False
    def __init__(self):
        self._types = {}
        self._custom = {}
        self._pspec_cache = {}

    def _add_type(self, prop_name, prop_type, pspec=None):
        if pspec:
            if gobject.type_is_a(pspec.value_type, gobject.TYPE_ENUM):
                prop_type.enum_class = pspec.enum_class
            elif gobject.type_is_a(pspec.value_type, gobject.TYPE_FLAGS):
                prop_type.flags_class = pspec.flags_class

            prop_type.label = pspec.nick
            prop_type.type = pspec.value_type
        else:
            if not prop_type.label:
                prop_type.label = prop_name.capitalize()

        if prop_name in self._types:
            msg = "There is already a property called %s registered"
            raise TypeError(msg % prop_name)

        self._types[prop_name] = prop_type

    def list(self, type_name, prop_registry):
        """
        List all the properties for type_name
        If parent is specified, child properties will also be included.
        This method will create the prop_types and save if they're
        not already created.

        @param type_name: name of type
        @type  type_name: string
        @param prop_registry: property registry
        @type  prop_registry: a PropRegistry
        """
        property_types = []
        for pspec in self._list_properties(type_name):
            prop = self._get_type_from_prop_name(type_name, pspec.name)
            if not prop:
                base = prop_registry._get_base_type(pspec.value_type)
                prop = self.create_type_from_pspec(pspec, base, type_name)
            property_types.append(prop)

        property_types.extend(self._list_custom(type_name))

        for prop_type in property_types[:]:
            if (not prop_type.readable or
                not prop_type.writable or
                not prop_type.enabled):
                property_types.remove(prop_type)

        return property_types

    def has_prop_type(self, type_name, prop_name):
        """
        True if a property from type_name called prop_name exists

        @param type_name: gtype name of owner
        @type  type_name: string
        @param prop_name: name of property
        @type  prop_name: string
        """
        prop_type = self._get_type_from_prop_name(type_name, prop_name)
        if not prop_type:
            prop_type = self._get_custom(type_name, prop_name)

        return prop_type

    def get_prop_type(self, type_name, prop_name, prop_registry):
        """
        Fetch a property from type_name called prop_name, it creates
        the property type if it's not already created.

        @param type_name: gtype name of owner
        @type  type_name: string
        @param prop_name: name of property
        @type  prop_name: string
        @param prop_registry: property registry
        @type  prop_registry: a PropRegistry
        """
        prop_type = self._get_type_from_prop_name(type_name, prop_name)
        if not prop_type:
            prop_type = self._get_custom(type_name, prop_name)

        if not prop_type:
            pspec = self.get_pspec(type_name, prop_name)
            if pspec:
                base = prop_registry._get_base_type(pspec.value_type)
                prop_type = self.create_type_from_pspec(pspec, base, type_name)

        return prop_type

    def add_custom_type(self, klass, type_name, prop_name, pspec):
        custom = self._custom.setdefault(type_name, [])
        if prop_name in custom:
            return
        klass.custom = True
        custom.append(prop_name)
        self._add_type(type_name + '::' + prop_name, klass, pspec)

    def _get_custom(self, type_name, prop_name):
        """
        Get custom properties from type_name, called prop_name
        """
        for prop_type in self._list_custom(type_name):
            if prop_type.name == prop_name:
                return prop_type

    def _list_custom(self, type_name):
        """
        List custom properties for type_name, check the type and
        all the parent types
        """
        retval = []
        while True:
            for prop_name in self._custom.get(type_name, []):
                full_name = type_name + '::' + prop_name
                retval.append(self._types[full_name])

            if type_name == 'GObject':
                break

            type_name = gobject.type_name(gobject.type_parent(type_name))

        return retval

    def get_pspec(self, type_name, prop_name):
        for pspec in self._list_properties(type_name):
            if pspec.name == prop_name:
                return pspec

    def has_property(self, type_name, prop_name):
        "Checks if  type_name has a property with name prop_name"
        return self.get_pspec(type_name, prop_name) != None

    def _list_properties(self, type_name):
        "List all properties for type_name, and use a simple cache"
        pspecs = self._pspec_cache.get(type_name, [])
        if pspecs:
            return pspecs

        try:
            if self.child:
                pspecs = gtk.container_class_list_child_properties(type_name)
            else:
                pspecs = gobject.list_properties(type_name)
        except TypeError:
            raise PropertyError("Unknown GType name: %s" % type_name)
        pspecs = list(pspecs)

        self._pspec_cache[type_name] = pspecs

        return pspecs

    def _get_type_from_prop_name(self, type_name, prop_name):
        """Get the type for type_name::prop_name. If we don't have a PropType
        for this type_name, check in its parent types"""
        assert type(type_name) == str

        while True:
            full_name = type_name + '::' + prop_name
            if full_name in list(self._types.keys()):
                return self._types[full_name]

            if type_name == 'GObject':
                break

            type_name = gobject.type_name(gobject.type_parent(type_name))

    def create_type_from_pspec(self, pspec, base, type_name=None):
        if not type_name:
            type_name = gobject.type_name(pspec.owner_type)

        readable = pspec.flags & gobject.PARAM_READABLE != 0
        writable = pspec.flags & gobject.PARAM_WRITABLE != 0
        prop_type = PropMeta.new(type_name,
                                 pspec.name,
                                 pspec.blurb,
                                 base,
                                 pspec.value_type,
                                 pspec.owner_type,
                                 readable=readable,
                                 writable=writable,
                                 child=self.child,
                                 label=pspec.nick)
        self._set_interval_from_pspec(prop_type, pspec)
        self._add_type(prop_type.__name__, prop_type, pspec)
        return prop_type

    def _set_interval_from_pspec(self, prop_type, pspec):
        if not hasattr(pspec, 'maximum'):
            return

        maximum = pspec.maximum
        if maximum == -1 and pspec.value_type == gobject.TYPE_UINT:
            maximum = sys.maxsize
        prop_type.maximum = maximum

        prop_type.minimum = pspec.minimum
        prop_type.defautl = pspec.default_value

class ChildPropStorage(PropStorage):
    child = True

class PropRegistry:
    def __init__(self):
        self._bases = {}

        self._normal_storage = PropStorage()
        self._child_storage = ChildPropStorage()

    def add_base_type(self, klass, gtype):
        if not issubclass(klass, PropType):
            raise TypeError("klass needs to be a subclass of PropType")

        if gtype in self._bases:
            raise PropertyError("base type %r is already defined" % gtype)
        elif klass.base_type != None:
            raise PropertyError("klass type %r is already registered to %r" %
                                (klass, klass.base_type))

        self._bases[gtype] = klass
        klass.base_type = gtype

    def override_property(self, full_name, klass):
        """
        Create a property for full_name, (type::prop_name)
        If the property exists, please use override_simple
        @param klass: property class
        @type  klass: subclass of PropType
        @param full_name: complete name of property
        @type  full_name: string (typename::prop_name)
        """
        self._override_property_internal(self._normal_storage,
                                         full_name, klass)

    def override_child_property(self, full_name, klass):
        """
        Create a child property for full_name, (type::prop_name)
        If the property exists, please use override_simple_child
        @param klass: property class
        @type  klass: subclass of PropType
        @param full_name: complete name of property
        @type  full_name: string (typename::prop_name)
        """
        self._override_property_internal(self._child_storage, full_name, klass)
        klass.child = True

    def _override_property_internal(self, storage, full_name, klass):
        if not issubclass(klass, PropType):
            raise TypeError("klass needs to be a subclass of PropType")

        type_name, prop_name = full_name.split('::')
        pspec = storage.get_pspec(type_name, prop_name)

        # If we can't find the property in the GType, it's a custom type
        if not storage.has_property(type_name, prop_name):
            storage.add_custom_type(klass, type_name, prop_name, pspec)
        else:
            raise TypeError("%s exists, use override_simple" % full_name)

        # Note, that at this point we need to setup the overriden class
        # since we're not going to call the type constructor, so
        # this needs to be in sync with PropMeta.new

        klass.name = prop_name
        klass.owner_name = type_name
        klass.owner_type = gobject.type_from_name(type_name)

        if not issubclass(klass.custom_editor, PropertyCustomEditor):
            raise TypeError(
                "custom_editor %r for property class %r needs to be a "
                "subclass of PropertyCustomEditor " %
                (klass.custom_editor, klass))

    def override_simple(self, full_name, base=None, **kwargs):
        """
        Same as override_property, but creates and registeres the class
        @param full_name: complete name of property
        @param base:      base class
        @param child:     child property
        @param kwargs namespace to use in class
        """
        self._override_simple_internal(self._normal_storage,
                                       full_name, base, **kwargs)

    def override_simple_child(self, full_name, base=None, **kwargs):
        """
        Same as override_child_property, but creates and registeres the class
        @param full_name: complete name of property
        @param base:      base class
        @param child:     child property
        @param kwargs namespace to use in class
        """
        self._override_simple_internal(self._child_storage,
                                       full_name, base, **kwargs)

    def _override_simple_internal(self, storage, full_name, base=None,
                                  **kwargs):
        type_name, prop_name = full_name.split('::')

        pspec = storage.get_pspec(type_name, prop_name)
        if not pspec:
            raise TypeError("You can only override existing properties, %s "
                            "is not a property" % full_name)

        # check if the type already exist
        if storage.has_prop_type(type_name, prop_name):
            prop_type = storage.get_prop_type(type_name, prop_name, self)
            # If a base is provided we don't modify the original proptype
            # but create a new one. This ensures the parent prop is not changed
            if base and prop_type == base:
                klass = type(full_name, (base,), kwargs)

                prop_type = storage.create_type_from_pspec(pspec,
                                                           klass,
                                                           type_name=type_name)
            else:
                # simply update the prop_type
                for key, value in list(kwargs.items()):
                    setattr(prop_type, key, value)

        else:
            if not base:
                base = self._get_base_type(pspec.value_type)

            klass = type(full_name, (base,), kwargs)

            prop_type = storage.create_type_from_pspec(pspec,
                                                       klass,
                                                       type_name=type_name)

        return prop_type

    def list(self, type_name, parent=None):
        """
        List all the properties for type_name
        If parent is specified, child properties will also be included.
        This method will create the prop_types if they're not already created.

        @param type_name: name of type
        @type  type_name: string
        @param parent:    parent type of object
        @type  parent:    GType
        """
        property_types = self._normal_storage.list(type_name, self)

        if parent:
            parent_type = gobject.type_name(parent)
            property_types.extend(self._child_storage.list(parent_type, self))

        return property_types

    def get_prop_type(self, type_name, prop_name, parent=None):
        """
        Fetch a property from type_name called prop_name. It creates
        the property type if it's not already created.

        @param type_name: gtype name of owner
        @type  type_name: string
        @param prop_name: name of property
        @type  prop_name: string
        """
        if parent:
            parent_type = gobject.type_name(parent)
            prop_type = self._child_storage.get_prop_type(parent_type,
                                                          prop_name, self)

        else:
            prop_type = self._normal_storage.get_prop_type(type_name,
                                                           prop_name, self)

        return prop_type

    def _get_base_type(self, value_type):
        if gobject.type_is_a(value_type, gobject.TYPE_OBJECT):
            base_type = gobject.TYPE_OBJECT
        elif gobject.type_is_a(value_type, gobject.TYPE_ENUM):
            base_type = gobject.TYPE_ENUM
        elif gobject.type_is_a(value_type, gobject.TYPE_FLAGS):
            base_type = gobject.TYPE_FLAGS
        elif gobject.type_is_a(value_type, gobject.TYPE_BOXED):
            base_type = gobject.TYPE_BOXED
        else:
            base_type = value_type

        base = self._bases.get(base_type)
        if not base:
            raise PropertyError("Can't find base type for %r" % base_type)
        return base


    def get_writable_properties(self, widget, child):
        """
        Lists all the writable properties, eg, the ones we'd like
        to have saved in a persistent format, eg glade files

        @param widget: a gtk.Widget
        @param child: True to list child properties, False to omit them
        """
        if child:
            parent_type = gobject.type_name(widget.get_parent())
            property_types = self._child_storage.list(parent_type, self)

        else:
            type_name = gobject.type_name(widget)
            property_types = self._normal_storage.list(type_name, self)

        properties = []
        # write the properties
        for prop_type in property_types:
            # don't save the name property since it is saved in the id
            # attribute of the tag
            if prop_type.name == 'name':
                continue

            # Do not save non-editable gobject properties, eg
            # GtkWindow::screen
            # hack?
            if (prop_type.base_type == gobject.GObject.__gtype__ and
                not prop_type.editable):
                continue

            if not prop_type.persistent:
                continue

            properties.append(prop_type)

        return properties

class PropMeta(type):
    """
    I am a metaclass used for Property types.
    Currently I only offer a __repr__ method, but there
    is also a class method called new which is used as
    a factory for types which has me as a metaclass.
    """

    #@classmethod
    def new(mcs, owner_name, name, description, base, value_type, owner_type,
            readable=True, writable=True, child=False, label=None):

        cls = mcs(owner_name + '::' + name, (base,), {})

        # This can not be overridden
        cls.name =  name
        cls.description = description
        cls.type = value_type
        cls.owner_name = owner_name
        cls.owner_type = owner_type

        # Be very careful here.
        # Do not overwrite anything, since we want this class
        # to be as transparent as possible.
        # Eg when doing:
        #
        # class Overridden(Custom, Int):
        #   ...
        # pr.override('some-prop', Foo)
        #
        # We want to have methods and attributes in the overridden to
        # to take precedence over the one in the transparent subclass
        # we're going to create, so check before assigning anything.

        if cls.readable is None:
            cls.readable = readable

        if cls.writable is None:
            cls.writable = writable

        if not cls.readable:
            getter = PropType._value_not_readable
        elif child:
            getter = base._child_get
        else:
            getter = base._get

        if not cls.writable:
            setter = PropType._value_not_writable
        elif child:
            setter = base._child_set
        else:
            setter = base._set

        if cls.value is None:
            cls.value = property(getter, setter)
        if not cls.set:
            cls.set = setter
        if not cls.get:
            cls.get = getter
        cls.child = child

        if not cls.label:
            # XXX: tooltip
            if not label:
                label = name.capitalize()
            cls.label = label

        return cls
    new = classmethod(new)

    def __repr__(mcs):
        child = ''
        if mcs.child:
            child = ' (child)'
        return '<PropType %s%s>' % (mcs.__name__, child)

class PropType(object, metaclass=PropMeta):
    """
    I am the base class for all properties.
    I wrap a GParamSpec (or custom property types), instances
    of me wrap a GObject property.

    @cvar name:         name of the property, eg expand
    @cvar description:  description of the property (it'll show in a tooltip)
    @cvar owner_name:   owner gobject name: GtkWindow
    @cvar type:         type, GType of property
    @cvar owner_type:   type of owner
    @cvar value:        value property
    @cvar readable:     can the value be retreived
    @cvar writable:     if it's possible to change values
    @cvar set:          set method
    @cvar get:          get method
    @cvar child:
    @cvar enabled:
    @cvar translatable: marked as translatable
    @cvar label:        property descriptor for set/get
    @cvar default:
    @cvar custom:
    @cvar editable:
    @cvar base_type:
    @cvar editor:        editor
    @cvar custom_editor: specialized editor
    @cvar persistent:    Should we save the property?
    @cvar priority:      Higher priorities appear before lower priorities in
                         the property editor
    @cvar has_custom_default: the prop has been customized with a default
    """

    name = None
    description = None
    owner_name = None
    type = None
    owner_type = gobject.TYPE_INVALID
    value = None
    readable = None
    writable = None
    set = None
    get = None
    child = None
    enabled = True
    translatable = False
    label = None
    default = None
    custom = False
    editable = True
    base_type = None
    editor = None
    custom_editor = PropertyCustomEditor
    persistent = True
    priority = 100
    has_custom_default = False

    def __init__(self, gadget):
        self._initial = None

        cls = type(self)
        assert cls.name != None
        assert cls.owner_name != None
        assert cls.owner_type != gobject.TYPE_INVALID

        self.gadget = gadget
        self._project = gadget.project
        gobj = gadget.widget

        # Assert that the owner is of correct type
        if self.child:
            test_object = gobj.get_parent()
        else:
            test_object = gobj
        if not gobject.type_is_a(test_object, self.owner_type):
            raise TypeError("Expected an object of type %s, got %r" % (
                gobject.type_name(self.owner_type), test_object))

        self.load()

    #@property
    def get_object(self):
        return self.gadget.widget
    object = property(get_object)

    def load(self):
        parent_name = None
        if self.child:
            parent_name = gobject.type_name(self.object.get_parent())

        # use the gtk value if this is an internal child
        if self.gadget.internal_name or self.gadget.maintain_gtk_properties:
            self._initial = self.gadget.adaptor.get_default_prop_value(
                self, parent_name)
            # we have to initialize _value from the gtk widget
            if self.custom:
                self._value = self._initial
            else:
                if self.child:
                    parent = self.object.get_parent()
                    self._value = parent.child_get_property(self.object,
                                                            self.name)
                else:
                    # FIXME: This is broken, but needed for
                    # GtkRadioButton::group,
                    # see tests.test_glade.GladeTest.test_radiobutton
                    try:
                        self._value = self.object.get_property(self.name)
                    except TypeError:
                        self._value = None
        else:
            if self.custom or self.has_custom_default:
                self._initial = self.default
                self.set(self.default)
            else:
                self._initial = self.gadget.adaptor.get_default_prop_value(
                    self, parent_name)

    def get_project(self):
        return self._project

    def get_object_name(self):
        if isinstance(self.object, gtk.Widget):
            return self.object.name

        return repr(self.object)

    def add_change_notify(self, function):
        if not callable(function):
            raise TypeError("function must be callable")

        if not custom:
            if self.child:
                signal_name = 'child-notify'
            else:
                signal_name = 'notify'
            self.object.connect('%s::%s' % (signal_name, self.name),
                                lambda self, obj, pspec: function(self))
    def _value_not_readable(self):
        raise PropertyError("%s not readable" % self.name)

    def _value_not_writable(self, value):
        raise PropertyError("%s not writable" % self.name)

    def _child_get(self):
        try:
            parent = self.object.get_parent()
            return parent.child_get_property(self.object, self.name)
        except TypeError as e:
            raise TypeError(
                "Failed to get child property from %r for %s "
                "(%s:%s): `%s'" % (parent, self.object,
                                   gobject.type_name(self.owner_type),
                                   self.name, e))

    def _child_set(self, value):
        try:
            parent = self.object.get_parent()
            parent.child_set_property(self.object,
                                      self.name,
                                      value)
        except TypeError as e:
            raise TypeError("Failed to set property %s:%s to %s: `%s'" % (
                gobject.type_name(self.owner_type), self.name, value, e))


    def _get(self):
        try:
            return self.object.get_property(self.name)
        except TypeError as e:
            raise TypeError("Error while getting %s:%s `%s'" % (
                gobject.type_name(self.owner_type), self.name, e))

    def _set(self, value):
        try:
            self.object.set_property(self.name, value)
        except TypeError as e:
            raise TypeError("Error while setting %s:%s to %s `%s'" % (
                gobject.type_name(self.owner_type), self.name, value, e))

    def get_default(self, gobj):
        if self.custom or self.has_custom_default:
            return self.default

        if self.child:
            parent = gobj.get_parent()
            return parent.child_get_property(gobj, self.name)

        return gobj.get_property(self.name)

    def is_translatable(self):
        """
        Is the property translatable?
        By default we're using a class variable, but in
        some cases we want to be able to check the instance
        state and check if we want translatation, eg GtkButton::label,
        in those cases, just override this method
        """
        return self.translatable

    def save(self):
        """
        Marshal the property into a string
        Return None if it already contains the default value
        """
        value = self.get()
        if not self.has_custom_default:
            if value == None or value == self._initial:
                return

        return str(value)

    def __repr__(self):
        return '<Property %s:%s value=%r>' % (self.owner_name, self.name,
                                              self.get())

# XXX: This is a really bad name since it conflicts with
#      the custom widget
class CustomProperty(PropType):
    readable = True
    writable = True
    def __init__(self, object):
        self._functions = []
        super(CustomProperty, self).__init__(object)

    def get(self):
        return self._get()

    def set(self, value):
        self._set(value)

    def add_change_notify(self, function):
        self._functions.append(function)

    def notify(self):
        for function in self._functions:
            function(self)

    value = property(lambda self: self.get(),
                     lambda self, value: self.set(value))

# This is of course a big hack, but it is for some reason
# required for epydoc to run
if environ.epydoc:
    class FakeRegistry:
        def add_base_type(self, *args, **kwargs):
            pass
        def override_property(self, *args, **kwargs):
            pass
        def override_simple(self, *args, **kwargs):
            pass
        def get_prop_type(self, *args, **kwargs):
            pass
    prop_registry = FakeRegistry()
    custom = object
else:
    prop_registry = PropRegistry()
    custom = CustomProperty

class CustomChildProperty(custom):
    child = True
    def get(self):
        return self._child_get()

    def set(self, value):
        self._child_set(value)

class TransparentProperty(custom):
    """
    A transparent property is a property which does not
    update the property of the GObject, instead it keeps track
    of the value using other means.
    For instance GtkWindow::modal is a transparent property, since
    we do not want to make the window we editing modal.
    """
    def __init__(self, *args):
        super(TransparentProperty, self).__init__(*args)
        self._value = self._initial

    def get(self):
        return self._value

    def set(self, value):
        self._value = value

class BooleanType(PropType):
    default = False
prop_registry.add_base_type(BooleanType, gobject.TYPE_BOOLEAN)

class PointerType(PropType):
    pass
prop_registry.add_base_type(PointerType, gobject.TYPE_POINTER)

class StringType(PropType):
    default = ''
    multiline = False
    translatable = True
    has_i18n_context = True
    i18n_comment = ''
prop_registry.add_base_type(StringType, gobject.TYPE_STRING)

class ObjectType(PropType):

    implements(IReferencable)

    editable = False

    def set(self, widget):
        from gazpacho.gadget import Gadget

        # remove old reference
        old_widget = self.get()
        if old_widget:
            old_gadget = Gadget.from_widget(old_widget)
            old_gadget.references.remove_referrer(self)

        # add new reference
        if widget:
            gadget = Gadget.from_widget(widget)
            gadget.references.add_referrer(self)

        #super(ObjectType, self).set(widget)
        self._set(widget)

    def save(self):
        value = self.get()
        if value:
            return value.get_name()

    # IReferencable

    def remove_reference(self, gadget):
        """
        See L{gazpacho.interfaces.IReferencable.remove_reference}
        """
        self.object.set_property(self.name, None)

    def add_reference(self, gadget):
        """
        See L{gazpacho.interfaces.IReferencable.add_reference}
        """
        self.object.set_property(self.name, gadget.widget)

prop_registry.add_base_type(ObjectType, gobject.TYPE_OBJECT)

#
# Numbers
#

class BaseNumericType(PropType):
    """
    Base type class used for all numeric types,
    The additional class variables here are used to create a
    GtkAlignment for the editors
    """
    default = 0
    # GtkAdjustment
    minimum = 0
    maximum = 0
    step_increment = 1
    page_increment = 10

    # SpinButton
    # acceleration factor
    climb_rate = 1

    # number of decimals
    digits = 0

    #@classmethod
    def get_adjustment(cls):
        #if not cls.minimum <= cls.default <= cls.maximum:
        #    raise TypeError(
        #        "default must be larger than minimum and lower "
        #        "than maximum (%r <= %r <= %r) for %s::%s (%r)" % (
        #        cls.minimum, cls.default, cls.maximum,
        #        gobject.type_name(cls.owner_type), cls.name, cls.type))

        return gtk.Adjustment(value=cls.default,
                              lower=cls.minimum,
                              upper=cls.maximum,
                              step_incr=cls.step_increment,
                              page_incr=cls.page_increment)
    get_adjustment = classmethod(get_adjustment)

class BaseFloatType(BaseNumericType):
    """
    Base float type, used by float and double
    """
    default = 0.0
    minimum = -10.0**308
    maximum = 10.0**308
    digits = 2
    def save(self):
        if not BaseNumericType.save(self):
            return
        # Fix for #326723
        # kiwi/datatypes.py uses %.12g here, but it seems to
        # be too precise, 0.1 -> 0.10000000149
        value = self.get()
        strvalue = str(value)
        if strvalue.endswith('.0'):
            return strvalue
        return '%.8g' % value

class IntType(BaseNumericType):
    minimum = -int(2**31-1)
    maximum = int(2**31-1)
prop_registry.add_base_type(IntType, gobject.TYPE_INT)

class UIntType(BaseNumericType):
    maximum = 2**32-1
prop_registry.add_base_type(UIntType,  gobject.TYPE_UINT)

class FloatType(BaseFloatType):
    pass
prop_registry.add_base_type(FloatType, gobject.TYPE_FLOAT)

class DoubleType(BaseFloatType):
    pass
prop_registry.add_base_type(DoubleType, gobject.TYPE_DOUBLE)


#
# Enum & Flags
#

class EnumType(PropType):
    enum_class = None

    #@classmethod
    def get_choices(cls):
        choices = []
        if not hasattr(cls.enum_class, '__enum_values__'):
            raise UnsupportedProperty

        for enum in list(cls.enum_class.__enum_values__.values()):
            choices.append((enum.value_nick, enum))
        choices.sort()
        return choices
    get_choices = classmethod(get_choices)

    def save(self):
        value = self.get()
        if value == self._initial:
            return

        return enum_to_string(value, enum=self.enum_class)

prop_registry.add_base_type(EnumType, gobject.TYPE_ENUM)

class FlagsType(PropType):
    flags_class = None

    #@classmethod
    def get_choices(cls):
        if not hasattr(cls.flags_class, '__flags_values__'):
            raise UnsupportedProperty

        flags = [(flag.first_value_name, mask)
                     for mask, flag in list(cls.flags_class.__flags_values__.items())
                         # we avoid composite flags
                         if len(flag.value_names) == 1]
        flags.sort()
        return flags
    get_choices = classmethod(get_choices)

    def save(self):
        value = self.get()
        if value == self._initial:
            return
        return flags_to_string(value, flags=self.flags_class)

prop_registry.add_base_type(FlagsType, gobject.TYPE_FLAGS)

class BoxedType(PropType):
    pass
prop_registry.add_base_type(BoxedType, gobject.TYPE_BOXED)

class CharType(PropType):
    default = '\0'
prop_registry.add_base_type(CharType, gobject.TYPE_CHAR)

class AdjustmentType(TransparentProperty):
    def save(self):
        adj = self.get()
        s = []
        for value in (adj.value, adj.lower, adj.upper,
                      adj.step_increment, adj.page_increment,
                      adj.page_size):
            # If it's a float which ends with .0, convert to an int,
            # so we can avoid saving unnecessary decimals
            if str(value).endswith('.0'):
                value = int(value)
            s.append(str(value))
        return ' '.join(s)
prop_registry.add_base_type(AdjustmentType, gtk.Adjustment)

class ProxiedProperty(TransparentProperty):
    target_name = None

    def get_target(self):
        raise NotImplementedError

    def get(self):
        target = self.get_target()
        return target.get_property(self.target_name)

    def set(self, value):
        target = self.get_target()
        return target.set_property(self.target_name, value)

# TODO
#   TYPE_INT64
#   TYPE_LONG
#   TYPE_UCHAR
#   TYPE_UINT64
#   TYPE_ULONG
#   TYPE_UNICHAR

# Property commands
class CommandSetProperty(Command):
    def __init__(self,  property, value):
        description = _('Setting %s of %s') % (
            property.name, property.get_object_name())
        Command.__init__(self, description)
        self._property = property
        self._value = value

    def execute(self):
        new_value = self._value
        # store the current value for undo
        self._value = self._property.value
        self._property.set(new_value)

        # TODO: this should not be needed
        # if the property is the name, we explicitily set the name of the
        # gadget to trigger the notify signal so several parts of the
        # interface get updated
        if self._property.name == 'name':
            self._property.object.notify('name')

    def unifies(self, other):
        if isinstance(other, CommandSetProperty):
            return self._property is other._property
        return False

    def collapse(self, other):
        self.description = other.description
        other.description = None

class CommandSetTranslatableProperty(CommandSetProperty):
    def __init__(self,  property, value, comment, translatable, has_context):
        CommandSetProperty.__init__(self, property, value)
        self._comment = comment
        self._translatable = translatable
        self._has_context = has_context

    def execute(self):
        new_comment = self._comment
        new_translatable = self._translatable
        new_has_context = self._has_context

        # store the current value for undo
        self._comment = self._property.i18n_comment
        self._translatable = self._property.translatable
        self._has_context = self._property.has_i18n_context

        CommandSetProperty.execute(self)

        self._property.translatable = new_translatable
        self._property.i18n_comment = new_comment
        self._property.has_i18n_context = new_has_context

if __name__ == '__main__':
    win = gtk.Window()
    box = gtk.HBox()
    win.add(box)
    button = gtk.Button()
    box.pack_start(button)

    from gazpacho.catalog import load_catalogs
    load_catalogs()

    for ptype in prop_registry.list('GtkButton', gtk.VBox):
        print(ptype, ptype.value)

    #for ptype in prop_registry.list('GtkHBox', gtk.Window):
    #    #if not ptype.child:
    #    #    continue
    #    prop = ptype(box, None)
    #    if prop.readable:
    #        print ptype, prop.value

    raise SystemExit




"""Database persistent administrative settings with defaults.

This code is a large fork of the excellent "dbsettings" code found at
http://code.google.com/p/django-values/

The items set here are intended to be changeable during runtime, and do not require a 
programmer to test or install.

Appropriate:  Your google code for adwords.
Inappropriate: The keyedcache timeout for the store.

"""

from functions import *
from models import *
from values import *


#!/usr/bin/env python
# Copyright (C) 2007-2018 Etienne Robillard <tkadm30@yandex.com>
# All rights reserved.
"""Generic fixture loading support independent from database driver.

To add support for a database backend, one must create a subclass
of the YAMLFixture class and override the ``create`` method. The
database driver must support fixtures to use the functions covered
in the YAMLFixture class.

Usage:

 >>> sampleData = YAMLFixture('/path/to/fixture.yaml')
 >>> sampleData.create(db) #NOTE: this is currently not implemented :)

"""
try:
    from io import open
except ImportError:
    pass

import os         # built-in
import posixpath  # built-in
import yaml       # requires pyyaml

try:
    # for JSON encoding support
    import demjson as jsonlib # demjson
except ImportError:
    import json # default/fallback json lib
    json_encoder = json.encoder
else:
    json_encoder = jsonlib.encode

__all__ = ['YAMLFixture',]

class YAMLFixture(object):
    """A iterable YAML fixture loader class.

    Can be used to load multiple data structures contained
    in the same YAML file.

    >>> fixture = YAMLFixture('apps.yaml')
    >>> app_conf = fixture.get(app_label='restapp', category='web')
    >>> for obj in fixture.dumps():
    >>>    print obj
    >>> fixture.tojson() # etc...
    """

    #db = None

    def __init__(self, path, **kwargs):
        self.initial_kwargs = kwargs
        self.path = os.path.abspath(path)
        
        if 'schema_index_root' in kwargs:
            self.root = kwargs['schema_index_root']

        self.initial_data = self.safe_load(path)

    def get_index_root(self):
        """return the current index root -> str """
        if 'root' in self.__dict__.keys():
            return str(self.root)
        return None
        #raise ValueError('schema_index_root not set!')
    index_root = property(get_index_root)

    def create(self, db, **kwargs):
        """
        Override this method to customize how yaml data should
        be saved in the db.
        """
        raise NotImplementedError

    def safe_load(self, pathname):
        """Safe load a YAML file using the yaml.load method."""
            
        try:
            initial_data = yaml.load(open(pathname, 'r'))
        except (IOError, OSError):
            raise
        else:
            return initial_data
    
    def __str__(self):
        return "<YAMLFixture path=%s>"%self.path

    cache = {}

    def __iter__(self):

        index_root = self.index_root
        # Attempt to create an iterator using data in self.initial_data, or
        # use the whole data set in case index_root is not usable.
        if index_root is not None and index_root in self.initial_data:
            iterable = iter(self.initial_data[self.index_root])
        else:
            # warning, deprecated stuff
            #print 'no usable index_root found, defaulting to slurp mode'
            iterable = iter(self.initial_data)

        return iterable

    def get(self, **kwargs):
        """
        Converts a list of params to a dict instance if matching
        the fixture item.

        """

        if not 'name' in kwargs:
            obj_id = str(hash(self))
        else:
            obj_id = kwargs['name']

        try:
            obj = self.cache[obj_id]
            #print "found obj in cache: %r" % obj_id
        except KeyError:
            obj = self.getdata(obj_id, **kwargs)
        
        if obj is not None:
            return obj
        else:    
            raise ValueError("no match found for %r" % str(self))
    
    def filter(self, **kwargs):
        """Return a sequence (list) of matching fixture items.
        
        """
        result = []
        for obj in self: # I'm a iterator
            for k, v in kwargs.iteritems():
                if obj[k].__contains__(v) or obj[k] == v:
                    result.append(obj)
                continue
        return result

    def getdata(self, obj_id, **kwargs):
        """
        Gets the cached object instance if available (dict), 
        otherwise do nothing.
        """
        for kw in self: # I'm a iterator
            #assert isinstance(kw, dict), 'kw must be a dict here'
            for k, v in kwargs.iteritems():
                if kw[k].__contains__(v) or kw[k] == v: 
                    # return the whole dictionary if the key
                    # is matching the corresponding value
                    self.cache[obj_id] = kw
                    return kw
                else:
                    continue    
        return
    
    def dumps(self):
        """Returns a list."""
        return [item for item in self]

    def tojson(self, filter_by={}, strict=False, escape_unicode=True):
        """Converts the list to json.
        
        ``filter_by``: If given, use that to selectively filter items in the
                       fixture
        """
        data = []
        if len(filter_by.keys()) >= 1:
            related_items = self.filter(**filter_by)
            data.extend(related_items)
        else:
            data.extend(self.dumps()) # default mode is a simple "dump all"
        
        return self.json_encode(data, strict=strict, escape_unicode=escape_unicode)

    def json_encode(self, value, **kwargs):
        """Encodes ``value`` in JSON format"""
        return json_encoder(value, **kwargs)



import os, sys

from notmm.utils.configparse import is_string
from notmm.utils.django_compat import get_callable
from notmm.utils.wsgilib import HTTPRequest
from notmm.utils.markup import FormWrapper
from test_support import WSGIControllerTestCase

class HTTPRequestTestCase(WSGIControllerTestCase):

    def test_post_required(self):
        client = self.client
        response = client.post('/')





# Copyright (C) 2004,2005 by SICEm S.L. and Imendio AB
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

from gazpacho.properties import prop_registry, CustomProperty, BooleanType

class HasResizeGripProp(CustomProperty, BooleanType):
    label = "Has Resize Grip"
    default = True
    def get(self):
        return self.object.get_has_resize_grip()

    def set(self, value):
        self.object.set_has_resize_grip(value)

prop_registry.override_simple('GtkStatusbar::has-resize-grip',
                              HasResizeGripProp)




#!/usr/bin/env python
#
"""Configuration options for template rendering and
parsing.

This class provides a compatibility bridge between customizable 
template backends and the Django ``RequestContext`` object.

TODO: Add support for Genshi templates.
"""
import os
from notmm.utils.template.interfaces import TemplateLoaderFactory
from notmm.utils.django_settings import SettingsProxy

from beaker.cache import CacheManager
from beaker.util import parse_cache_config_options
cache_opts = {
    'cache.data_dir': '/tmp/cache/data',
    'cache.lock_dir': '/tmp/cache/lock',
    'cache.regions': 'short_term, long_term',
    'cache.short_term.type': 'ext:memcached',
    'cache.short_term.url': '127.0.0.1:11211',
    'cache.short_term.expire': '1200',
    'cache.long_term.type': 'ext:memcached',
    'cache.long_term.url' : '127.0.0.1:11211',
    'cache.long_term.expire': '86400'
}

__all__ = ('get_template_loader', 'get_app_conf')



class TemplateException(Exception):
    pass

def get_app_conf(name='development.ini'):

    try:
        from notmm.utils.configparse import loadconf
        rootdir = os.environ['ROOTDIR']
        app_conf = loadconf(name)
    except (ImportError, KeyError):
        app_conf = {}
    return app_conf

# register a default template backend instance 
def get_template_loader(cache_enabled=True):
    
    _settings = SettingsProxy(autoload=True).get_settings()
    _cache_enabled = getattr(_settings, 'ENABLE_BEAKER', cache_enabled)
    
    TemplateLoaderFactory.configure(
        #backend='mako', 
        #template_loader_class='CachedTemplateLoader',
        kwargs={
            'directories' : _settings.TEMPLATE_DIRS, 
            'cache_enabled':_cache_enabled,
            'cache_args': {
              'manager': CacheManager(**parse_cache_config_options(cache_opts))
              }
            }
    )
    loader = TemplateLoaderFactory.get_loader()
    return loader




from django.contrib import admin
from django.utils.translation import ugettext_lazy as _
from product.admin import ProductOptions
from satchmo_ext.brand.models import Brand, BrandTranslation, BrandCategory, BrandCategoryTranslation, BrandProduct, BrandCategoryProduct
from product.models import Product
from satchmo_utils.thumbnail.field import ImageWithThumbnailField
from satchmo_utils.thumbnail.widgets import AdminImageWithThumbnailWidget

class BrandTranslation_Inline(admin.StackedInline):
    model = BrandTranslation
    extra = 1
    verbose_name = _("Translation")
    verbose_name_plural = _("Translations")
    
    formfield_overrides = {
        ImageWithThumbnailField : {'widget' : AdminImageWithThumbnailWidget},
    }
    

class BrandCategoryTranslation_Inline(admin.StackedInline):
    model = BrandCategoryTranslation
    extra = 1
    verbose_name = _("Translation")
    verbose_name_plural = _("Translations")

    formfield_overrides = {
        ImageWithThumbnailField : {'widget' : AdminImageWithThumbnailWidget},
    }

class BrandCategoryTranslationOptions(admin.ModelAdmin):
    fieldsets = (
        (None, {'fields' : ('brandcategory', 'languagecode', 'name', 'description', 'picture')}),
    )

class BrandOptions(admin.ModelAdmin):
    inlines = [BrandTranslation_Inline]

    
class BrandProductInline(admin.TabularInline):
    model = BrandProduct
    max_num = 1
    verbose_name = _("Brand")
    verbose_name_plural = _("Brands")
    

class BrandTranslationOptions(admin.ModelAdmin):
    fieldsets = (
        (None, {'fields' : ('brand', 'languagecode', 'name', 'description', 'picture')}),
    )

class BrandCategoryOptions(admin.ModelAdmin):
    inlines = [BrandCategoryTranslation_Inline]


class BrandCategoryInline(admin.TabularInline):
    model = BrandCategoryProduct
    extra = 1
    verbose_name = _("Brand Category")
    verbose_name_plural = _("Brand Categories")

admin.site.register(Brand, BrandOptions)
admin.site.register(BrandCategory, BrandCategoryOptions)
admin.site.unregister(Product)
ProductOptions.inlines.append(BrandProductInline)
ProductOptions.inlines.append(BrandCategoryInline)
admin.site.register(Product, ProductOptions)



"""
URLConf for Satchmo Contacts.
"""

from django.conf.urls.defaults import *
from signals_ahoy.signals import collect_urls
from satchmo_store import contact

urlpatterns = patterns('satchmo_store.contact.views',
    (r'^$', 'view', {}, 'satchmo_account_info'),
    (r'^update/$', 'update', {}, 'satchmo_profile_update'), 
)

collect_urls.send(sender=contact, patterns=urlpatterns)


from django.utils.translation import ugettext_lazy as _
from livesettings import *

TAX_MODULE = config_get('TAX', 'MODULE')
TAX_MODULE.add_choice(('tax.modules.us_sst', _('USA: Streamlined Sales Tax')))
TAX_GROUP = config_get_group('TAX')

# varies state-to-state, so we need to support this being variable.
config_register(
     BooleanValue(TAX_GROUP,
         'TAX_SHIPPING',
         description=_("Tax Shipping in ANY jurisdiction?"),
         requires=TAX_MODULE,
         requiresvalue='tax.modules.us_sst',
         default=True)
)

config_register(
     StringValue(TAX_GROUP,
         'TAX_CLASS',
         description=_("TaxClass for shipping"),
         help_text=_("Select a TaxClass that should be applied for shipments."),
         #TODO: [BJK] make this dynamic - doesn't work to have it be preloaded.
         default='Shipping'
     )
)



from django import template
from django.contrib.sites.models import Site
from django.core import urlresolvers
from livesettings import config_value, SettingNotSet
from satchmo_utils import url_join
import logging

log = logging.getLogger('configuration.config_tags')

register = template.Library()

def force_space(value, chars=40):
    """Forces spaces every `chars` in value"""

    chars = int(chars)
    if len(value) < chars:
        return value
    else:    
        out = []
        start = 0
        end = 0
        looping = True
        
        while looping:
            start = end
            end += chars
            out.append(value[start:end])
            looping = end < len(value)
    
    return ' '.join(out)

def break_at(value,  chars=40):
    """Force spaces into long lines which don't have spaces"""
    
    chars = int(chars)
    value = unicode(value)
    if len(value) < chars:
        return value
    else:
        out = []
        line = value.split(' ')
        for word in line:
            if len(word) > chars:
                out.append(force_space(word, chars))
            else:
                out.append(word)
                
    return " ".join(out)

register.filter('break_at', break_at)

def config_boolean(option):
    """Looks up the configuration option, returning true or false."""
    args = option.split('.')
    try:
        val = config_value(*args)
    except:
        log.warn('config_boolean tag: Tried to look up config setting "%s", got SettingNotSet, returning False', option)
        val = False
    if val:
        return "true"
    else:
        return ""

register.filter('config_boolean', config_boolean)

def admin_site_views(view):
    """Returns a formatted list of sites, rendering for view, if any"""

    if view:
        path = urlresolvers.reverse(view)
    else:
        path = None

    links = []
    for site in Site.objects.all():
        paths = ["http://", site.domain]
        if path:
            paths.append(path)
            
        links.append((site.name, url_join(paths)))
    
    ret = {
        'links' : links,
    }
    return ret
    

register.inclusion_tag('livesettings/_admin_site_views.html')(admin_site_views)



from decimal import Decimal
from django.contrib.sites.models import Site
from django.test import TestCase
from product.models import Option, OptionGroup, Product, Price
from product.modules.configurable.models import ConfigurableProduct, ProductVariation
import datetime
import keyedcache
from product.utils import serialize_options, productvariation_details


class OptionUtilsTest(TestCase):
    """Test the utilities used for serialization of options and selected option details."""
    fixtures = ['products.yaml']

    def test_base_sort_order(self):
        p = Product.objects.get(slug='dj-rocks')
        serialized = serialize_options(p.configurableproduct)
        self.assert_(len(serialized), 2)
        self.assertEqual(serialized[0]['id'], 1)
        got_vals = [opt.value for opt in serialized[0]['items']]
        self.assertEqual(got_vals, ['S','M','L'])
        self.assertEqual(serialized[1]['id'], 2)

    def test_reordered(self):
        p = Product.objects.get(slug='dj-rocks')

        pv = p.configurableproduct.productvariation_set.all()[0]
        orig_key = pv.optionkey
        orig_detl = productvariation_details(p, False, None, create=True)

        sizegroup = OptionGroup.objects.get(name="sizes")
        sizegroup.sort_order = 100
        sizegroup.save()

        # reverse ordering
        for opt in sizegroup.option_set.all():
            opt.sort_order = 100-opt.sort_order
            opt.save()

        serialized = serialize_options(p.configurableproduct)
        self.assert_(len(serialized), 2)
        self.assertEqual(serialized[1]['id'], 1)
        got_vals = [opt.value for opt in serialized[1]['items']]
        self.assertEqual(got_vals, ['L','M','S'])

        pv2 = ProductVariation.objects.get(pk=pv.pk)
        self.assertEqual(orig_key, pv2.optionkey)
        reorder_detl = productvariation_details(p, False, None)
        self.assertEqual(orig_detl, reorder_detl)


class OptionGroupTest(TestCase):

    def setUp(self):
        self.site=Site.objects.get_current()
        sizes = OptionGroup.objects.create(name="sizes", sort_order=1, site=self.site)
        option_small = Option.objects.create(option_group=sizes, name="Small", value="small", sort_order=1)
        option_large = Option.objects.create(option_group=sizes, name="Large", value="large", sort_order=2, price_change=1)
        colors = OptionGroup.objects.create(name="colors", sort_order=2, site=self.site)
        option_black = Option.objects.create(option_group=colors, name="Black", value="black", sort_order=1)
        option_white = Option.objects.create(option_group=colors, name="White", value="white", sort_order=2, price_change=3)

        # Change an option
        option_white.price_change = 5
        option_white.sort_order = 2
        option_white.save()

        self.sizes = sizes
        self.option_small = option_small
        self.option_large = option_large
        self.colors = colors
        self.option_black = option_black
        self.option_white = option_white

    def testConfigurable(self):
        """Create a configurable product, testing ordering and price"""
        django_shirt = Product.objects.create(slug="django-shirt", name="Django shirt", site=self.site)
        shirt_price = Price.objects.create(product=django_shirt, price="10.5")
        django_config = ConfigurableProduct.objects.create(product=django_shirt)
        django_config.option_group.add(self.sizes, self.colors)
        ordering = django_config.option_group.order_by('name')
        self.assertEqual(ordering[0], self.colors)
        self.assertEqual(ordering[1], self.sizes)
        django_config.save()

        # Create a product variation
        white_shirt = Product.objects.create(slug="django-shirt_small_white",
            name="Django Shirt (White/Small)", site=self.site)
        pv_white = ProductVariation.objects.create(product=white_shirt, parent=django_config)
        pv_white.options.add(self.option_white, self.option_small)
        self.assertEqual(pv_white.unit_price, Decimal("15.50"))

    def testConfigurableSlugs(self):
        """Create a product with a slug that could conflict with an
        automatically generated product's slug."""

        django_shirt = Product.objects.create(slug="django-shirt", name="Django shirt", site=self.site)
        shirt_price = Price.objects.create(product=django_shirt, price="10.5")
        django_config = ConfigurableProduct.objects.create(product=django_shirt)
        django_config.option_group.add(self.sizes, self.colors)
        django_config.save()

        # Create a product with a slug that could conflict with an automatically
        # generated product's slug.
        clash_shirt = Product.objects.create(slug="django-shirt_small_black",
            name="Django Shirt (Black/Small)", site=self.site)

        # Automatically create the rest of the product variations
        django_config.create_subs = True
        django_config.save()
        self.assertEqual(ProductVariation.objects.filter(parent=django_config).count(), 4)


class ProductTest(TestCase):
    """Test Product functions"""
    fixtures = ['l10n-data.yaml','sample-store-data.yaml', 'products.yaml', 'test-config.yaml']

    def tearDown(self):
        keyedcache.cache_delete()

    def test_quantity_price_productvariation(self):
        """Check quantity price for a productvariation"""

        # base product
        product = Product.objects.get(slug='dj-rocks')
        self.assertEqual(product.unit_price, Decimal("20.00"))
        self.assertEqual(product.unit_price, product.get_qty_price(Decimal('1')))

        # product with no price delta
        product = Product.objects.get(slug='dj-rocks-s-b')
        self.assertEqual(product.unit_price, Decimal("20.00"))
        self.assertEqual(product.unit_price, product.get_qty_price(Decimal('1')))

        # product which costs more due to details
        product = Product.objects.get(slug='dj-rocks-l-bl')
        self.assertEqual(product.unit_price, Decimal("23.00"))
        self.assertEqual(product.unit_price, product.get_qty_price(Decimal('1')))


    def test_quantity_price_productvariation_expiring(self):
        """Check expiring quantity price for a productvariation"""

        # base product
        product = Product.objects.get(slug='dj-rocks')
        self.assertEqual(product.unit_price, Decimal("20.00"))
        self.assertEqual(product.unit_price, product.get_qty_price(Decimal('1')))

        today = datetime.datetime.now()
        aweek = datetime.timedelta(days=7)
        nextwk = today + aweek

        # new price should override the old one
        price = Price.objects.create(product=product, quantity=Decimal('1'), price=Decimal("10.00"), expires=nextwk)
        self.assertEqual(product.unit_price, Decimal("10.00"))

        # product with no price delta
        product = Product.objects.get(slug='dj-rocks-s-b')
        self.assertEqual(product.unit_price, Decimal("10.00"))

        # product which costs more due to details
        product = Product.objects.get(slug='dj-rocks-l-bl')
        self.assertEqual(product.unit_price, Decimal("13.00"))

        # now test explicit expiring price on a product variation
        pvprice = Price.objects.create(product=product, quantity=Decimal('1'), price=Decimal("5.00"), expires=nextwk)
        self.assertEqual(product.unit_price, Decimal("5.00"))

class ConfigurableProductTest(TestCase):
    """Test ConfigurableProduct."""
    fixtures = ['products.yaml']

    def tearDown(self):
        keyedcache.cache_delete()

    def test_get_variations_for_options(self):
        # Retrieve the objects.
        dj_rocks = ConfigurableProduct.objects.get(product__slug="dj-rocks")
        option_small = Option.objects.get(pk=1)
        option_black = Option.objects.get(pk=4)
        option_hard_cover = Option.objects.get(pk=7)

        # Test filtering for one valid option.
        self.assertEqual([variation.pk for variation in
            dj_rocks.get_variations_for_options([option_small])], [6, 7, 8])
        # Test filtering for two valid options.
        self.assertEqual([variation.pk for variation in
            dj_rocks.get_variations_for_options([option_small, option_black])],
            [6])
        # Test filtering for an option that cannot apply to the product.
        self.assertEqual(
            len(dj_rocks.get_variations_for_options([option_hard_cover])), 0)
        # Test filtering for nothing.
        self.assertEqual([variation.pk for variation in
            dj_rocks.get_variations_for_options([])],
            [6, 7, 8, 9, 10, 11, 12, 13, 14])


if __name__ == "__main__":
    import doctest
    doctest.testmod()




"""Adds a new HTTPS handler to urllib2, which uses the SSL library from Python 2.6."""

import logging
log = logging.getLogger('sslurllib')
import sys

# We only need this wrapper for Python versions < 2.6
# sys.hexversion is guaranteed to always increment
if sys.hexversion >= 0x020600F0: 
    runningPython26 = True
else: 
    runningPython26 = False
    
try:
    import ssl
    _sane = True
except:
    log.warning('ssl is not installed, please install it from http://pypi.python.org/pypi/ssl/')
    _sane = False
    
if _sane and not runningPython26:
    import httplib, socket, urllib2
    __all__ = ['HTTPSv2Connection', 'HTTPSv2Handler']
    
    log.debug('Installing SSLv2 HTTPS Methods into urllib2')
    
    class HTTPSv2Connection(httplib.HTTPConnection):
        """This class allows communication via SSLv2."""

        default_port = httplib.HTTPS_PORT

        def __init__(self, host, port=None, key_file=None, cert_file=None,
                     strict=None):
            httplib.HTTPConnection.__init__(self, host, port, strict)
            self.key_file = key_file
            self.cert_file = cert_file

        def connect(self):
            "Connect to a host on a given (SSL) port."
            log.debug('HTTPSv2 connecting to %s:%s', self.host, self.port)
            
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            ssl_sock = ssl.wrap_socket(sock,
                                       ca_certs=self.cert_file)

            ssl_sock.connect((self.host, self.port))
            self.sock = ssl_sock

    class HTTPSv2Handler(urllib2.HTTPSHandler):
        """Overrides the base urllib2 HTTPSHandler."""
        
        def https_open(self, req):
            return self.do_open(HTTPSv2Connection, req)

    # now instantiate the HTTPSv2Handler and install it.
    v2handler = HTTPSv2Handler()
    opener = urllib2.build_opener(v2handler)
    
    # this will make our new subclassed HTTPSHandler be used for all HTTPSConnections
    urllib2.install_opener(opener)
    



#!/usr/bin/env python
# Copyright (c) 2007-2019 Etienne Robillard
# All rights reserved.
"""
Markdown2 utils

"""
import sys
PY3K = sys.version_info[0] == 3

if PY3K:
    from ._markdown3 import Markdown, MarkdownError
else:
    from ._markdown2 import Markdown, MarkdownError

__all__ = ['convert2markdown']

   
def convert2markdown(value, ispath=False):
    """Transform a markdown string into valid HTML markup. 
    
    ``ispath`` -- Indicates that ``value`` should be treated as a Unix pathname.
    """

    try:
        if ispath is not False:
            with open(value, 'r') as f:
                value = f.read()
                f.close()
        # Convert the string to html
        md = Markdown()
        html_obj = md.convert(value)
    except MarkdownError as e:
        # Error converting the document to html
        raise e
    else:
        return html_obj

#convert_to_markdown = convert2markdown



from notmm.utils.urlmap import RegexURLMap

urlpatterns = RegexURLMap()
urlpatterns.add_routes('', 
        #(r'^login/$', 'sandbox.views.login', {
        #    'template_name' : 'sandbox/login.mako'
        #}),

        # with_session decorator test
        (r'^test$', 'sandbox.views.test_session', {
            'template_name' : 'sandbox/test.mako'
        }),
        # with_schevo_database test
        (r'^test.schevo$', 'sandbox.views.test_schevo_database', {
            'template_name' : 'sandbox/test.mako'
        }),
        # handle302 test (Temporary redirects)
        (r'^moved/$', 'sandbox.handlers.handle302', {
            'location' : '/index.html'
        }),
        # handle500 test (sethandle)
        (r'^croak$', 'sandbox.views.croak')
)
urlpatterns.commit()




#!/usr/bin/env python

# Modified version of Django's make-messages.py
# This is from Django revision 6445.
# It has been modified to parse ".txt" and ".rml" files in addition to its
# normal behavior of parsing ".py" and ".html" files.

# Need to ensure that the i18n framework is enabled
from django.conf import settings
settings.configure(USE_I18N = True)

from django.utils.translation import templatize
import re
import os
import sys
import getopt
from itertools import dropwhile

pythonize_re = re.compile(r'\n\s*//')

def make_messages():
    localedir = None

    if os.path.isdir(os.path.join('conf', 'locale')):
        localedir = os.path.abspath(os.path.join('conf', 'locale'))
    elif os.path.isdir('locale'):
        localedir = os.path.abspath('locale')
    else:
        print "This script should be run from the django svn tree or your project or app tree."
        print "If you did indeed run it from the svn checkout or your project or application,"
        print "maybe you are just missing the conf/locale (in the django tree) or locale (for project"
        print "and application) directory?"
        print "make-messages.py doesn't create it automatically, you have to create it by hand if"
        print "you want to enable i18n for your project or application."
        sys.exit(1)

    (opts, args) = getopt.getopt(sys.argv[1:], 'l:d:va')

    lang = None
    domain = 'django'
    verbose = False
    all = False

    for o, v in opts:
        if o == '-l':
            lang = v
        elif o == '-d':
            domain = v
        elif o == '-v':
            verbose = True
        elif o == '-a':
            all = True

    if domain not in ('django', 'djangojs'):
        print "currently make-messages.py only supports domains 'django' and 'djangojs'"
        sys.exit(1)
    if (lang is None and not all) or domain is None:
        print "usage: make-messages.py -l <language>"
        print "   or: make-messages.py -a"
        sys.exit(1)

    languages = []

    if lang is not None:
        languages.append(lang)
    elif all:
        languages = [el for el in os.listdir(localedir) if not el.startswith('.')]

    for lang in languages:

        print "processing language", lang
        basedir = os.path.join(localedir, lang, 'LC_MESSAGES')
        if not os.path.isdir(basedir):
            os.makedirs(basedir)

        pofile = os.path.join(basedir, '%s.po' % domain)
        potfile = os.path.join(basedir, '%s.pot' % domain)

        if os.path.exists(potfile):
            os.unlink(potfile)

        all_files = []
        for (dirpath, dirnames, filenames) in os.walk("."):
            all_files.extend([(dirpath, f) for f in filenames])
        all_files.sort()
        for dirpath, file in all_files:
            if domain == 'djangojs' and file.endswith('.js'):
                if verbose: sys.stdout.write('processing file %s in %s\n' % (file, dirpath))
                src = open(os.path.join(dirpath, file), "rb").read()
                src = pythonize_re.sub('\n#', src)
                open(os.path.join(dirpath, '%s.py' % file), "wb").write(src)
                thefile = '%s.py' % file
                cmd = 'xgettext %s -d %s -L Perl --keyword=gettext_noop --keyword=gettext_lazy --keyword=ngettext_lazy:1,2 --from-code UTF-8 -o - "%s"' % (
                    os.path.exists(potfile) and '--omit-header' or '', domain, os.path.join(dirpath, thefile))
                (stdin, stdout, stderr) = os.popen3(cmd, 't')
                msgs = stdout.read()
                errors = stderr.read()
                if errors:
                    print "errors happened while running xgettext on %s" % file
                    print errors
                    sys.exit(8)
                old = '#: '+os.path.join(dirpath, thefile)[2:]
                new = '#: '+os.path.join(dirpath, file)[2:]
                msgs = msgs.replace(old, new)
                if msgs:
                    open(potfile, 'ab').write(msgs)
                os.unlink(os.path.join(dirpath, thefile))
            elif domain == 'django' and (file.endswith('.py') or file.endswith('.html') or file.endswith('.txt') or file.endswith('.rml')):
                thefile = file
                if file.endswith('.html') or file.endswith('.txt') or file.endswith('.rml'):
                    src = open(os.path.join(dirpath, file), "rb").read()
                    thefile = '%s.py' % file
                    open(os.path.join(dirpath, thefile), "wb").write(templatize(src))
                if verbose:
                    sys.stdout.write('processing file %s in %s\n' % (file, dirpath))
                cmd = 'xgettext -d %s -L Python --keyword=gettext_noop --keyword=gettext_lazy --keyword=ngettext_lazy:1,2 --keyword=ugettext_noop --keyword=ugettext_lazy --keyword=ungettext_lazy:1,2 --from-code UTF-8 -o - "%s"' % (
                    domain, os.path.join(dirpath, thefile))
                (stdin, stdout, stderr) = os.popen3(cmd, 't')
                msgs = stdout.read()
                errors = stderr.read()
                if errors:
                    print "errors happened while running xgettext on %s" % file
                    print errors
                    sys.exit(8)
                if thefile != file:
                    old = '#: '+os.path.join(dirpath, thefile)[2:]
                    new = '#: '+os.path.join(dirpath, file)[2:]
                    msgs = msgs.replace(old, new)
                if os.path.exists(potfile):
                    # Strip the header
                    msgs = '\n'.join(dropwhile(len, msgs.split('\n')))
                else:
                    msgs = msgs.replace('charset=CHARSET', 'charset=UTF-8')
                if msgs:
                    open(potfile, 'ab').write(msgs)
                if thefile != file:
                    os.unlink(os.path.join(dirpath, thefile))

        if os.path.exists(potfile):
            (stdin, stdout, stderr) = os.popen3('msguniq --to-code=utf-8 "%s"' % potfile, 'b')
            msgs = stdout.read()
            errors = stderr.read()
            if errors:
                print "errors happened while running msguniq"
                print errors
                sys.exit(8)
            open(potfile, 'w').write(msgs)
            if os.path.exists(pofile):
                (stdin, stdout, stderr) = os.popen3('msgmerge -q "%s" "%s"' % (pofile, potfile), 'b')
                msgs = stdout.read()
                errors = stderr.read()
                if errors:
                    print "errors happened while running msgmerge"
                    print errors
                    sys.exit(8)
            open(pofile, 'wb').write(msgs)
            os.unlink(potfile)

if __name__ == "__main__":
    make_messages()



from datetime import datetime
from decimal import Decimal
from django.contrib.sites.models import Site
from django.contrib.sites.models import Site
from django.db import models
from django.utils.translation import ugettext_lazy as _
from l10n.utils import moneyfmt
from livesettings import config_value
from payment.modules.giftcertificate.utils import generate_certificate_code
from payment.utils import get_processor_by_key
from product.models import Product
from satchmo_store.contact.models import Contact
from satchmo_store.shop.models import OrderPayment, Order
import logging

GIFTCODE_KEY = 'GIFTCODE'
log = logging.getLogger('giftcertificate.models')

class GiftCertificateManager(models.Manager):

    def from_order(self, order):
        code = order.get_variable(GIFTCODE_KEY, "")
        log.debug("GiftCert.from_order code=%s", code)
        if code:
            site = order.site
            return GiftCertificate.objects.get(code__exact=code.value, valid__exact=True, site=site)
        raise GiftCertificate.DoesNotExist()

class GiftCertificate(models.Model):
    """A Gift Cert which holds value."""
    site = models.ForeignKey(Site, null=True, blank=True, verbose_name=_('Site'))
    order = models.ForeignKey(Order, null=True, blank=True, related_name="giftcertificates", verbose_name=_('Order'))
    code = models.CharField(_('Certificate Code'), max_length=100,
        blank=True, null=True)
    purchased_by =  models.ForeignKey(Contact, verbose_name=_('Purchased by'),
        blank=True, null=True, related_name='giftcertificates_purchased')
    date_added = models.DateField(_("Date added"), null=True, blank=True)
    valid = models.BooleanField(_('Valid'), default=True)
    message = models.CharField(_('Message'), blank=True, null=True, max_length=255)
    recipient_email = models.EmailField(_("Email"), blank=True, max_length=75)
    start_balance = models.DecimalField(_("Starting Balance"), decimal_places=2,
        max_digits=8)

    objects = GiftCertificateManager()

    def balance(self):
        b = Decimal(self.start_balance)
        for usage in self.usages.all():
            log.info('usage: %s' % usage)
            b = b - Decimal(usage.balance_used)

        return b

    balance = property(balance)

    def apply_to_order(self, order):
        """Apply up to the full amount of the balance of this cert to the order.

        Returns new balance.
        """
        amount = min(order.balance, self.balance)
        log.info('applying %s from giftcert #%i [%s] to order #%i [%s]', 
            moneyfmt(amount), 
            self.id, 
            moneyfmt(self.balance), 
            order.id, 
            moneyfmt(order.balance))
            
        processor = get_processor_by_key('PAYMENT_GIFTCERTIFICATE')
        orderpayment = processor.record_payment(order=order, amount=amount)
        self.orderpayment = orderpayment
        return self.use(amount, orderpayment=orderpayment)

    def use(self, amount, notes="", orderpayment=None):
        """Use some amount of the gift cert, returning the current balance."""
        u = GiftCertificateUsage(notes=notes, balance_used = amount,
            orderpayment=orderpayment, giftcertificate=self)
        u.save()
        return self.balance

    def save(self, **kwargs):
        if not self.pk:
            self.date_added = datetime.now()
        if not self.code:
            self.code = generate_certificate_code()
        if not self.site:
            self.site = Site.objects.get_current()
        super(GiftCertificate, self).save(**kwargs)

    def __unicode__(self):
        sb = moneyfmt(self.start_balance)
        b = moneyfmt(self.balance)
        return u"Gift Cert: %s/%s" % (sb, b)

    class Meta:
        verbose_name = _("Gift Certificate")
        verbose_name_plural = _("Gift Certificates")

class GiftCertificateUsage(models.Model):
    """Any usage of a Gift Cert is logged with one of these objects."""
    usage_date = models.DateField(_("Date of usage"), null=True, blank=True)
    notes = models.TextField(_('Notes'), blank=True, null=True)
    balance_used = models.DecimalField(_("Amount Used"), decimal_places=2,
        max_digits=8, )
    orderpayment = models.ForeignKey(OrderPayment, null=True, verbose_name=_('Order Payment'))
    used_by = models.ForeignKey(Contact, verbose_name=_('Used by'),
        blank=True, null=True, related_name='giftcertificates_used')
    giftcertificate = models.ForeignKey(GiftCertificate, related_name='usages')

    def __unicode__(self):
        return u"GiftCertificateUsage: %s" % self.balance_used

    def save(self, **kwargs):
        if not self.pk:
            self.usage_date = datetime.now()
        super(GiftCertificateUsage, self).save(**kwargs)


class GiftCertificateProduct(models.Model):
    """
    The product model for a Gift Certificate
    """
    product = models.OneToOneField(Product, verbose_name=_('Product'), primary_key=True)
    is_shippable = False
    discountable = False

    def __unicode__(self):
        return u"GiftCertificateProduct: %s" % self.product.name
        
    def _get_subtype(self):
        return 'GiftCertificateProduct'        

    def order_success(self, order, order_item):
        log.debug("Order success called, creating gift certs on order: %s", order)
        message = ""
        email = ""
        for detl in order_item.orderitemdetail_set.all():
            if detl.name == "email":
                email = detl.value
            elif detl.name == "message":
                message = detl.value

        price=order_item.line_item_price
        log.debug("Creating gc for %s", price)
        gc = GiftCertificate(
            order = order,
            start_balance= price,
            purchased_by = order.contact,
            valid=True,
            message=message,
            recipient_email=email
            )
        gc.save()
    
    def save(self, **kwargs):
        if hasattr(self.product,'_sub_types'):
            del self.product._sub_types
        super(GiftCertificateProduct, self).save(**kwargs)

    class Meta:
        verbose_name = _("Gift certificate product")
        verbose_name_plural = _("Gift certificate products")

import config
PAYMENT_PROCESSOR=True


from l10n.models import Country, AdminArea
from django.contrib import admin
from django.utils.translation import get_language, ugettext_lazy as _


class AdminArea_Inline(admin.TabularInline):
    model = AdminArea
    extra = 1

class CountryOptions(admin.ModelAdmin):
    
    def make_active(self, request, queryset):
        rows_updated = queryset.update(active=True)
        if rows_updated == 1:
            message_bit = _("1 country was")
        else:
            message_bit = _("%s countries were" % rows_updated)
        self.message_user(request, _("%s successfully marked as active") % message_bit)
    make_active.short_description = _("Mark selected countries as active")
    
    def make_inactive(self, request, queryset):
        rows_updated = queryset.update(active=False)
        if rows_updated == 1:
            message_bit = _("1 country was")
        else:
            message_bit = _("%s countries were" % rows_updated)
        self.message_user(request, _("%s successfully marked as inactive") % message_bit)
    make_inactive.short_description = _("Mark selected countries as inactive")
    
    list_display = ('printable_name', 'iso2_code','active')
    list_filter = ('continent', 'active')
    search_fields = ('name', 'iso2_code', 'iso3_code')
    actions = ('make_active', 'make_inactive')
    inlines = [AdminArea_Inline]

admin.site.register(Country, CountryOptions)




# Copyright (C) 2004,2005 by SICEm S.L. and Imendio AB
#               2005,2006 Async Open Source
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

from gazpacho.properties import prop_registry, CustomProperty, StringType
from gazpacho.propertyeditor import EditorPropertyImage, EditorPropertyStock

class FileProp(CustomProperty, StringType):
    # we can remove this class when depending in gtk 2.8 since this property
    # is readable there
    editable = True
    translatable = False

    editor = EditorPropertyImage

    # We need to override get_default, since
    # the standard mechanism to get the initial value
    # is calling get_property, but ::file is not marked as readable
    def get_default(self, gobj):
        return ""

    def load(self):
        value = super(FileProp, self).load()
        self._initial = value
        self.set(value)

    def get(self):
        return self.object.get_data('image-file-name')

    def set(self, value):
        if value:
            self.object.set_data('image-file-name', value)
            self.object.set_property('file', value)

    def save(self):
        return self.get()

prop_registry.override_simple('GtkImage::file', FileProp)

class StockProp(CustomProperty, StringType):
    default = "gtk-missing-image"
    translatable = False
    editor = EditorPropertyStock
    def save(self):
        if self.object.get_data('image-file-name'):
            return
        return super(StockProp, self).save()

prop_registry.override_simple('GtkImage::stock', StockProp)



"""Simplified ORM abstraction module for Schevo""" 

from ._databaseproxy import DatabaseProxy, ConnectionError
from ._relation import RelationProxy
from .schevo_compat import XdserverProxy
from .zodb_compat import ClientStorageProxy

__all__ = [
    'DatabaseProxy',
    'RelationProxy', 
    'XdserverProxy',
    'ConnectionError',
    #'models', 
    'schevo_compat',
    'zodb_compat'
    #'AnonymousUser',
    #'scoped_session',
    #'with_session',
    #'with_schevo_database']
    ]    



"""From http://code.activestate.com/recipes/279004/"""
import re

# Build up the regex patterns piece by piece
POSTAL_ZONES = ['AB', 'AL', 'B' , 'BA', 'BB', 'BD', 'BH', 'BL', 'BN', 'BR',
                'BS', 'BT', 'CA', 'CB', 'CF', 'CH', 'CM', 'CO', 'CR', 'CT',
                'CV', 'CW', 'DA', 'DD', 'DE', 'DG', 'DH', 'DL', 'DN', 'DT',
                'DY', 'E' , 'EC', 'EH', 'EN', 'EX', 'FK', 'FY', 'G' , 'GL',
                'GY', 'GU', 'HA', 'HD', 'HG', 'HP', 'HR', 'HS', 'HU', 'HX',
                'IG', 'IM', 'IP', 'IV', 'JE', 'KA', 'KT', 'KW', 'KY', 'L' ,
                'LA', 'LD', 'LE', 'LL', 'LN', 'LS', 'LU', 'M' , 'ME', 'MK',
                'ML', 'N' , 'NE', 'NG', 'NN', 'NP', 'NR', 'NW', 'OL', 'OX',
                'PA', 'PE', 'PH', 'PL', 'PO', 'PR', 'RG', 'RH', 'RM', 'S' ,
                'SA', 'SE', 'SG', 'SK', 'SL', 'SM', 'SN', 'SO', 'SP', 'SR',
                'SS', 'ST', 'SW', 'SY', 'TA', 'TD', 'TF', 'TN', 'TQ', 'TR',
                'TS', 'TW', 'UB', 'W' , 'WA', 'WC', 'WD', 'WF', 'WN', 'WR',
                'WS', 'WV', 'YO', 'ZE']
POSTAL_ZONES_ONE_CHAR = [zone for zone in POSTAL_ZONES if len(zone) == 1]
POSTAL_ZONES_TWO_CHARS = [zone for zone in POSTAL_ZONES if len(zone) == 2]
THIRD_POS_CHARS = ['A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'J', 'K', 'S',
                   'T', 'U', 'W']
FOURTH_POS_CHARS = ['A', 'B', 'E', 'H', 'M', 'N', 'P', 'R', 'V', 'W', 'X',
                    'Y']
INCODE_CHARS = ['A', 'B', 'D', 'E', 'F', 'G', 'H', 'J', 'L', 'N', 'P', 'Q',
                'R', 'S', 'T', 'U', 'W', 'X', 'Y', 'Z']
OUTCODE_PATTERN = (r'(' + 
                   r'(?:(?:' +
                   '|'.join(POSTAL_ZONES_ONE_CHAR) +
                   r')(?:\d[' +
                   ''.join(THIRD_POS_CHARS) +
                   r']|\d{1,2}))' +
                   r'|' +
                   r'(?:(?:' +
                   '|'.join(POSTAL_ZONES_TWO_CHARS) +
                   r')(?:\d[' +
                   ''.join(FOURTH_POS_CHARS) +
                   r']|\d{1,2}))' +
                   r')')
INCODE_PATTERN = (r'(\d[' +
                  ''.join(INCODE_CHARS) +
                  r'][' +
                  ''.join(INCODE_CHARS) +
                  r'])')
POSTCODE_PATTERN = OUTCODE_PATTERN + INCODE_PATTERN
STANDALONE_OUTCODE_PATTERN = OUTCODE_PATTERN + r'\s*$'

# Compile regexs
POSTCODE_REGEX = re.compile(POSTCODE_PATTERN)
STANDALONE_OUTCODE_REGEX = re.compile(STANDALONE_OUTCODE_PATTERN)

def parse_uk_postcode(postcode, strict=True, incode_mandatory=True):
    postcode = postcode.replace(' ', '').upper() # Normalize
    
    if len(postcode) > 7:
        raise ValueError('Postcode too long')
    
    # Validate postcode
    if strict:
        
        # Try for full postcode match
        postcode_match = POSTCODE_REGEX.match(postcode)
        if postcode_match:
            return postcode_match.group(1, 2)
            
        # Try for outcode only match
        outcode_match = STANDALONE_OUTCODE_REGEX.match(postcode)
        if outcode_match:
            if incode_mandatory:
                raise ValueError('Incode mandatory')
            else:
                return outcode_match.group(1), ''

        # Try Girobank special case
        if postcode == 'GIR0AA':
            return 'GIR', '0AA'
        elif postcode == 'GIR':
            if incode_mandatory:
                raise ValueError('Incode mandatory')
            else:
                return 'GIR', ''
        
        # None of the above
        raise ValueError('Invalid postcode')
        
    # Just chop up whatever we've been given.
    else:
        # Outcode only
        if len(postcode) <= 4:
            if incode_mandatory:
                raise ValueError('Incode mandatory')
            else:
                return postcode, ''
        # Full postcode
        else:
            return postcode[:-3], postcode[-3:]



#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

import os

from kiwi.environ import Library

dirname = os.path.abspath(
    os.path.join(os.path.dirname(os.path.realpath(__file__)), '..'))

lib = Library('gazpacho', root=dirname)
if lib.uninstalled:
    lib.add_global_resource('catalog', 'catalogs')
    lib.add_global_resource('glade', 'glade')
    lib.add_global_resource('resource', 'resources')
    lib.add_global_resource('doc', '.')
    lib.add_global_resource('pixmap', 'pixmaps')
    lib.add_global_resource('pixmap', 'pixmaps/kiwi')
    lib.add_global_resource('plugins', 'plugins')
lib.enable_translation()

__version__ = "0.7.2"




from django import template
from django.db import models
register = template.Library()

try:
    ''.rsplit
    def rsplit(s, delim, maxsplit):
        return s.rsplit(delim, maxsplit)
except AttributeError:
    def rsplit(s, delim, maxsplit):
        """
        Return a list of the words of the string s, scanning s
        from the end. To all intents and purposes, the resulting
        list of words is the same as returned by split(), except
        when the optional third argument maxsplit is explicitly
        specified and nonzero. When maxsplit is nonzero, at most
        maxsplit number of splits - the rightmost ones - occur,
        and the remainder of the string is returned as the first
        element of the list (thus, the list will have at most
        maxsplit+1 elements). New in version 2.4.
        >>> rsplit('foo.bar.baz', '.', 0)
        ['foo.bar.baz']
        >>> rsplit('foo.bar.baz', '.', 1)
        ['foo.bar', 'baz']
        >>> rsplit('foo.bar.baz', '.', 2)
        ['foo', 'bar', 'baz']
        >>> rsplit('foo.bar.baz', '.', 99)
        ['foo', 'bar', 'baz']
        """
        assert maxsplit >= 0
           
        if maxsplit == 0: return [s]
           
        # the following lines perform the function, but inefficiently.
        #  This may be adequate for compatibility purposes
        items = s.split(delim)
        if maxsplit < len(items):
            items[:-maxsplit] = [delim.join(items[:-maxsplit])]
        return items

class FilterAdminApplistNode(template.Node):
    def __init__(self, listname, varname):
        self.listname = listname
        self.varname = varname

    def render(self, context):
        all_apps = {}
        for app in models.get_apps():
            name = len(rsplit(app.__name__, '.', 0))>1 and rsplit(app.__name__, '.', 0)[-2] or app.__name__
            all_apps[name] = app.__name__

        filtered_app_list = []
        for entry in context[self.listname]:
            app = all_apps.get(entry['name'].lower(),'')
            if not app.startswith('satchmo_'):
                filtered_app_list.append(entry)
        context[self.varname] = filtered_app_list
        return ''

def filter_admin_app_list(parser, token):
    """Filters the list of installed apps returned by
       django.contrib.admin.templatetags.adminapplist,
       excluding apps installed by satchmo.
    """
    tokens = token.contents.split()
    if len(tokens) < 4:
        raise template.TemplateSyntaxError, "'%s' tag requires two arguments" % tokens[0]
    if tokens[2] != 'as':
        raise template.TemplateSyntaxError, "Second argument to '%s' tag must be 'as'" % tokens[0]
    return FilterAdminApplistNode(tokens[1], tokens[3])

register.tag('filter_admin_app_list', filter_admin_app_list)
